##
# File:      PdbxBuild.py
# Orignal:   2009-Oct-13  Jdw
#
# Updates:
# 12-Mar-2012  jdw  Create full parent relations  methods in dictionary API.
#                     Provide isEnumerated test.
# 12-Aug-2012  jdw  Add alternative alternate metadata attribute method placeholders
# 30-Aug-2012  jdw  Implementation of alternate metadata.
# 21-Sep-2012  jdw  Adjust the presententation and sorting in dump and print methods.
# 23-Oct-2012  jdw  Adjust path and reorganize
# 10-Mar-2014  jdw  class DictionaryApi(object) in this module is deprecated -- Please use
#                   this class from dictionary/DictionaryApi()
##
"""
Utility classes for PDBx/mmCIF dictionary and data file management.
"""

from __future__ import absolute_import
from __future__ import print_function
from six.moves import zip
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import sys
from pdbx_v2.reader.PdbxReader import PdbxReader
from pdbx_v2.writer.PdbxWriter import PdbxWriter
from pdbx_v2.reader.PdbxContainers import *
from pdbx_v2.reader.DataCategory import DataCategory


class MethodDefinition(object):

    def __init__(self, method_id, code='calculate', language='Python', inline=None):
        self.method_id = method_id
        self.language = language
        self.code = code
        self.inline = inline

    def getId(self):
        return self.method_id

    def getLanguage(self):
        return self.language

    def getInline(self):
        return self.inline

    def printIt(self, fh=sys.stderr):
        fh.write("------------- Method definition -------------\n")
        fh.write("Id:           %s\n" % self.method_id)
        fh.write("Code:         %s\n" % self.code)
        fh.write("Language:     %s\n" % str(self.language))
        fh.write("Inline text:  %s\n" % str(self.inline))


class MethodReference(object):

    def __init__(self, method_id, type='attribute', category=None, attribute=None):
        self.method_id = method_id
        self.type = type
        self.categoryName = category
        self.attributeName = attribute

    def getId(self):
        return self.method_id

    def getType(self):
        return self.type

    def getCategoryName(self):
        return self.categoryName

    def getAttributeName(self):
        return self.attributeName

    def printIt(self, fh=sys.stderr):
        fh.write("--------------- Method Reference -----------------\n")
        fh.write("Id:             %s\n" % self.method_id)
        fh.write("Type:           %s\n" % self.type)
        fh.write("Category name:  %s\n" % str(self.categoryName))
        fh.write("Attribute name: %s\n" % str(self.attributeName))


class DictionaryApi(object):

    def __init__(self, dictObjList, verbose=False):
        self.dictObjList = dictObjList
        self.fullIndex = {}

        # self.dict={}
        # for d in self.dictObjList:
        #    if (not self.dict.has_key(d.getName())):
        #        self.dict[d.getName()]=[]
        #    self.dict[d.getName()].append(d)
        #
        #
        # Map category name to the unique list of attributes
        self.catNameIndex = {}
        #
        # Map dictionary objects names to definition containers -
        self.definitionIndex = {}
        #
        # data section/objects of the dictionary by category name -
        self.dataIndex = {}
        #
        # Map of types id->(regex,primitive_type)
        self.typesDict = {}
        #
        self.enumD = {'ENUMERATION_VALUE': ('item_enumeration', 'value'),
                      'ENUMERATION_DETAIL': ('item_enumeration', 'detail'),
                      'ENUMERATION_TUPLE': ('item_enumeration', None),
                      'ITEM_LINKED_PARENT': ('item_linked', 'parent_name'),
                      'ITEM_LINKED_CHILD': ('item_linked', 'child_name'),
                      'DATA_TYPE_CODE': ('item_type', 'code'),
                      'DATA_TYPE_REGEX': ('item_type_list', 'construct'),
                      'DATA_TYPE_PRIMITIVE': ('item_type_list', 'primitive_code'),
                      'ITEM_MANDATORY_CODE': ('item', 'mandatory_code'),
                      'ITEM_DESCRIPTION': ('item_description', 'description'),
                      'ITEM_UNITS': ('item_units', 'code'),
                      'ITEM_MANDATORY_CODE': ('item', 'mandatory_code'),
                      'ITEM_DEFAULT_VALUE': ('item_default', 'value'),

                      'ITEM_EXAMPLE_CASE': ('item_examples', 'case'),
                      'ITEM_EXAMPLE_DETAIL': ('item_examples', 'detail'),
                      'ITEM_RANGE_MAXIMUM': ('item_range', 'maximum'),
                      'ITEM_RANGE_MINIMUM': ('item_range', 'minimum'),
                      'CATEGORY_KEY_ITEMS': ('category_key', 'name'),
                      'CATEGORY_EXAMPLE_CASE': ('category_examples', 'case'),
                      'CATEGORY_EXAMPLE_DETAIL': ('category_examples', 'detail'),
                      'CATEGORY_MANDATORY_CODE': ('category', 'mandatory_code'),
                      'CATEGORY_DESCRIPTION': ('category', 'description'),
                      #
                      'DATA_TYPE_CODE_NDB': ('ndb_item_type', 'code'),
                      'ITEM_DESCRIPTION_NDB': ('ndb_item_description', 'description'),
                      'ENUMERATION_VALUE_NDB': ('ndb_item_enumeration', 'value'),
                      'ENUMERATION_DETAIL_NDB': ('ndb_item_enumeration', 'detail'),
                      'ITEM_MANDATORY_CODE_NDB': ('ndb_item', 'mandatory_code'),
                      'ITEM_EXAMPLE_CASE_NDB': ('ndb_item_examples', 'case'),
                      'ITEM_EXAMPLE_DETAIL_NDB': ('ndb_item_examples', 'detail'),
                      'ITEM_RANGE_MAXIMUM_NDB': ('ndb_item_range', 'maximum'),
                      'ITEM_RANGE_MINIMUM_NDB': ('ndb_item_range', 'minimum'),
                      'CATEGORY_EXAMPLE_CASE_NDB': ('ndb_category_examples', 'case'),
                      'CATEGORY_EXAMPLE_DETAIL_NDB': ('ndb_category_examples', 'detail'),
                      'CATEGORY_DESCRIPTION_NDB': ('ndb_category_description', 'description'),
                      #
                      'DATA_TYPE_CODE_PDBX': ('pdbx_item_type', 'code'),
                      'ITEM_DESCRIPTION_PDBX': ('pdbx_item_description', 'description'),
                      'ENUMERATION_VALUE_PDBX': ('pdbx_item_enumeration', 'value'),
                      'ENUMERATION_DETAIL_PDBX': ('pdbx_item_enumeration', 'detail'),
                      'ITEM_MANDATORY_CODE_PDBX': ('pdbx_item', 'mandatory_code'),
                      'ITEM_EXAMPLE_CASE_PDBX': ('pdbx_item_examples', 'case'),
                      'ITEM_EXAMPLE_DETAIL_PDBX': ('pdbx_item_examples', 'detail'),
                      'ITEM_RANGE_MAXIMUM_PDBX': ('pdbx_item_range', 'maximum'),
                      'ITEM_RANGE_MINIMUM_PDBX': ('pdbx_item_range', 'minimum'),
                      'CATEGORY_EXAMPLE_CASE_PDBX': ('pdbx_category_examples', 'case'),
                      'CATEGORY_EXAMPLE_DETAIL_PDBX': ('pdbx_category_examples', 'detail'),
                      'CATEGORY_DESCRIPTION_PDBX': ('pdbx_category_description', 'description'),
                      #
                      'CATEGORY_CONTEXT': ('pdbx_category_context', 'type'),
                      'ITEM_CONTEXT': ('pdbx_item_context', 'type'),
                      'ENUMERATION_CLOSED_FLAG': ('pdbx_item_enumeration_details', 'closed_flag')
                      }
        #
        self.methodDict = {}
        self.methodIndex = {}
        #
        self.__getIndex()
        self.__getMethods()
        self.__getTypes()

    def definitionExists(self, definitionName):
        if definitionName in self.definitionIndex:
            return True
        return False

    def getTypeCode(self, category, attribute):
        return self.__get('DATA_TYPE_CODE', category, attribute)

    def getTypeCodeAlt(self, category, attribute, fallBack=True):
        v = self.getTypeCodePdbx(category, attribute)
        if v is None:
            v = self.getTypeCodeNdb(category, attribute)
        if fallBack and v is None:
            v = self.getTypeCode(category, attribute)
        return v

    def getTypeCodeNdb(self, category, attribute):
        return self.__get('DATA_TYPE_CODE_NDB', category, attribute)

    def getTypeCodePdbx(self, category, attribute):
        return self.__get('DATA_TYPE_CODE_PDBX', category, attribute)

    def getDefaultValue(self, category, attribute):
        return self.__get('ITEM_DEFAULT_VALUE', category, attribute)

    def getMandatoryCode(self, category, attribute):
        return self.__get('ITEM_MANDATORY_CODE', category, attribute)

    def getMandatoryCodeAlt(self, category, attribute, fallBack=True):
        v = self.getMandatoryCodePdbx(category, attribute)
        if v is None:
            v = self.getMandatoryCodeNdb(category, attribute)
        if fallBack and v is None:
            v = self.getMandatoryCode(category, attribute)
        return v

    def getMandatoryCodeNdb(self, category, attribute):
        return self.__get('ITEM_MANDATORY_CODE_NDB', category, attribute)

    def getMandatoryCodePdbx(self, category, attribute):
        return self.__get('ITEM_MANDATORY_CODE_PDBX', category, attribute)

    def getTypeRegex(self, category, attribute):
        code = self.__get('DATA_TYPE_CODE', category, attribute)
        if code in self.typesDict:
            return self.typesDict[code][1]
        return None

    def getTypeRegexAlt(self, category, attribute, fallBack=True):
        v = self.getTypeRegexPdbx(category, attribute)
        if v is None:
            v = self.getTypeRegexNdb(category, attribute)
        if fallBack and v is None:
            v = self.getTypeRegex(category, attribute)
        return v

    def getTypeRegexNdb(self, category, attribute):
        code = self.__get('DATA_TYPE_CODE_NDB', category, attribute)
        if code in self.typesDict:
            return self.typesDict[code][1]
        return None

    def getTypeRegexPdbx(self, category, attribute):
        code = self.__get('DATA_TYPE_CODE_PDBX', category, attribute)
        if code in self.typesDict:
            return self.typesDict[code][1]
        return None

    def getTypePrimitive(self, category, attribute):
        code = self.__get('DATA_TYPE_CODE', category, attribute)
        if code in self.typesDict:
            return self.typesDict[code][0]
        return None

    def getContextList(self, category, attribute):
        return self.__getList('ITEM_CONTEXT', category, attribute)

    def getCategoryContextList(self, category):
        return self.__getList('CATEGORY_CONTEXT', category, attribute=None)

    def getEnumList(self, category, attribute):
        return self.__getList('ENUMERATION_VALUE', category, attribute)

    def getEnumListAlt(self, category, attribute, fallBack=True):
        vL = self.getEnumListPdbx(category, attribute)
        if len(vL) < 1:
            vL = self.getEnumListNdb(category, attribute)
        if fallBack and len(vL) < 1:
            vL = self.getEnumList(category, attribute)
        return vL

    def getEnumListNdb(self, category, attribute):
        return self.__getList('ENUMERATION_VALUE_NDB', category, attribute)

    def getEnumListPdbx(self, category, attribute):
        return self.__getList('ENUMERATION_VALUE_PDBX', category, attribute)

    def isEnumerated(self, category, attribute):
        return (len(self.__getList('ENUMERATION_VALUE', category, attribute)) > 0)

    def isEnumeratedAlt(self, category, attribute, fallBack=True):
        eC = len(self.__getList('ENUMERATION_VALUE_PDBX', category, attribute))
        if (eC == 0):
            eC = len(self.__getList('ENUMERATION_VALUE_NDB', category, attribute))
        if (fallBack and (eC == 0)):
            eC = len(self.__getList('ENUMERATION_VALUE', category, attribute))
        return (eC > 0)

    def getEnumerationClosedFlag(self, category, attribute):
        return self.__get('ENUMERATION_CLOSED_FLAG', category, attribute)

    def getParentList(self, category, attribute):
        return self.__getList('ITEM_LINKED_PARENT', category, attribute)

    def getChildList(self, category, attribute):
        return self.__getList('ITEM_LINKED_CHILD', category, attribute)

    def getUnits(self, category, attribute):
        return self.__get('ITEM_UNITS', category, attribute)

    def getImplicitList(self):
        iL = []
        for name, dL in self.definitionIndex.items():
            for d in dL:
                type = d.getType()
                if (type == "definition" and d.isAttribute()):
                    catN = CifName.categoryPart(name)
                    attN = CifName.attributePart(name)
                    if (self.__get('ITEM_MANDATORY_CODE', catN, attN) == 'implicit'):
                        if name not in iL:
                            iL.append(name)
        return iL

    def getDescription(self, category, attribute):
        return self.__get('ITEM_DESCRIPTION', category, attribute)

    def getDescriptionAlt(self, category, attribute, fallBack=True):
        v = self.getDescriptionPdbx(category, attribute)
        if v is None:
            v = self.getDescriptionNdb(category, attribute)
        if fallBack and v is None:
            v = self.getDescription(category, attribute)
        return v

    def getDescriptionNdb(self, category, attribute):
        return self.__get('ITEM_DESCRIPTION_NDB', category, attribute)

    def getDescriptionPdbx(self, category, attribute):
        return self.__get('ITEM_DESCRIPTION_PDBX', category, attribute)

    def getExampleList(self, category, attribute):
        exCL = self.__getListAll('ITEM_EXAMPLE_CASE', category, attribute)
        exDL = self.__getListAll('ITEM_EXAMPLE_DETAIL', category, attribute)
        exL = []
        for exC, exD in zip(exCL, exDL):
            exL.append((exC, exD))
        return exL

    def getExampleListAlt(self, category, attribute, fallBack=True):
        vL = self.getExampleListPdbx(category, attribute)
        if len(vL) < 1:
            vL = self.getExampleListNdb(category, attribute)
        if fallBack and len(vL) < 1:
            vL = self.getExampleList(category, attribute)
        return vL

    def getExampleListNdb(self, category, attribute):
        exCL = self.__getListAll('ITEM_EXAMPLE_CASE_NDB', category, attribute)
        exDL = self.__getListAll('ITEM_EXAMPLE_DETAIL_NDB', category, attribute)
        exL = []
        for exC, exD in zip(exCL, exDL):
            exL.append((exC, exD))
        return exL

    def getExampleListPdbx(self, category, attribute):
        exCL = self.__getListAll('ITEM_EXAMPLE_CASE_PDBX', category, attribute)
        exDL = self.__getListAll('ITEM_EXAMPLE_DETAIL_PDBX', category, attribute)
        exL = []
        for exC, exD in zip(exCL, exDL):
            exL.append((exC, exD))
        return exL

    def getBoundaryList(self, category, attribute):
        minL = self.__getListAll('ITEM_RANGE_MINIMUM', category, attribute)
        maxL = self.__getListAll('ITEM_RANGE_MAXIMUM', category, attribute)
        bL = []
        for vMin, vMax in zip(minL, maxL):
            bL.append((vMin, vMax))
        return bL

    def getBoundaryListAlt(self, category, attribute, fallBack=True):
        vL = self.getBoundaryListPdbx(category, attribute)
        if len(vL) < 1:
            vL = self.getBoundaryListNdb(category, attribute)
        if fallBack and len(vL) < 1:
            vL = self.getBoundaryList(category, attribute)
        return vL

    def getBoundaryListNdb(self, category, attribute):
        minL = self.__getListAll('ITEM_RANGE_MINIMUM_NDB', category, attribute)
        maxL = self.__getListAll('ITEM_RANGE_MAXIMUM_NDB', category, attribute)
        bL = []
        for vMin, vMax in zip(minL, maxL):
            bL.append((vMin, vMax))
        #
        return bL

    def getBoundaryListPdbx(self, category, attribute):
        minL = self.__getListAll('ITEM_RANGE_MINIMUM_PDBX', category, attribute)
        maxL = self.__getListAll('ITEM_RANGE_MAXIMUM_PDBX', category, attribute)
        bL = []
        for vMin, vMax in zip(minL, maxL):
            bL.append((vMin, vMax))
        #
        return bL

    def getCategoryKeyList(self, category):
        return self.__getList('CATEGORY_KEY_ITEMS', category, attribute=None)

    def getCategoryMandatoryCode(self, category):
        return self.__get('CATEGORY_MANDATORY_CODE', category, attribute=None)

    def getCategoryDescription(self, category):
        return self.__get('CATEGORY_DESCRIPTION', category, attribute=None)

    def getCategoryDescriptionAlt(self, category, fallBack=True):
        v = self.getCategoryDescriptionPdbx(category)
        if v is None:
            v = self.getCategoryDescriptionNdb(category)
        if (fallBack and v is None):
            v = self.getCategoryDescription(category)
        return v

    def getCategoryDescriptionNdb(self, category):
        val = self.__get('CATEGORY_DESCRIPTION_NDB', category, attribute=None)
        return val

    def getCategoryDescriptionPdbx(self, category):
        val = self.__get('CATEGORY_DESCRIPTION_PDBX', category, attribute=None)
        return val

    def getCategoryExampleList(self, category):
        exCL = self.__getListAll('CATEGORY_EXAMPLE_CASE', category, attribute=None)
        exDL = self.__getListAll('CATEGORY_EXAMPLE_DETAIL', category, attribute=None)
        exL = []
        for exC, exD in zip(exCL, exDL):
            exL.append((exC, exD))
        return exL

    def getCategoryExampleListAlt(self, category, fallBack=True):
        vL = self.getCategoryExampleListPdbx(category)
        if len(vL) < 1:
            vL = self.getCategoryExampleListNdb(category)
        if fallBack and len(vL) < 1:
            vL = self.getCategoryExampleList(category)
        return vL

    def getCategoryExampleListNdb(self, category):
        exCL = self.__getListAll('CATEGORY_EXAMPLE_CASE_NDB', category, attribute=None)
        exDL = self.__getListAll('CATEGORY_EXAMPLE_DETAIL_NDB', category, attribute=None)
        exL = []
        for exC, exD in zip(exCL, exDL):
            exL.append((exC, exD))
        return exL

    def getCategoryExampleListPdbx(self, category):
        exCL = self.__getListAll('CATEGORY_EXAMPLE_CASE_PDBX', category, attribute=None)
        exDL = self.__getListAll('CATEGORY_EXAMPLE_DETAIL_PDBX', category, attribute=None)
        exL = []
        for exC, exD in zip(exCL, exDL):
            exL.append((exC, exD))
        return exL

    def getParentDictionary(self):
        """ Create a dictionary of parents relations accross all definnitions

            as d[child]=[parent,parent,...]

            Exclude self parents.
        """
        parentD = {}
        pAtN = self.enumD['ITEM_LINKED_PARENT'][1]
        cAtN = self.enumD['ITEM_LINKED_CHILD'][1]

        for dObj in self.dictObjList:
            dc = dObj.getObj(self.enumD['ITEM_LINKED_PARENT'][0])
            if dc is not None:
                idxP = dc.getIndex(pAtN)
                idxC = dc.getIndex(cAtN)
                for row in dc.getRowList():
                    pVal = row[idxP]
                    cVal = row[idxC]
                    if pVal == cVal:
                        continue
                    if cVal not in parentD:
                        parentD[cVal] = []
                    parentD[cVal].append(pVal)
        #
        return parentD

    def dumpEnumFeatures(self, fh=sys.stdout):
        for k, vL in self.catNameIndex.items():
            uvL = list(set(vL))
            for v in sorted(uvL):
                itL = self.getEnumList(k, v)
                if len(itL) > 0:
                    fh.write("-----------------------------------------------\n")
                    fh.write("       Category : %s\n" % k)
                    fh.write("       Attribute: %s\n" % v)
                    fh.write("     Description: \n%s\n" % self.getDescription(k, v))
                    fh.write("            Type: %s\n" % self.getTypeCode(k, v))
                    fh.write("  Primitive type: %s\n" % self.getTypePrimitive(k, v))
                    fh.write("      Regex type: %s\n" % self.getTypeRegex(k, v))
                    fh.write("      Enum list length %d\n" % len(itL))
                    for it in itL:
                        fh.write("      Enum: %s\n" % it)

    def dumpFeatures(self, fh=sys.stdout):
        for k, vL in self.catNameIndex.items():
            uvL = list(set(vL))
            fh.write("-----------------------------------------------\n")
            fh.write("Category: %s has %d attributes\n" % (k, len(uvL)))
            fh.write("     Category description: %s\n" % self.getCategoryDescription(k))
            fh.write(" Alt category description: %s\n" % self.getCategoryDescriptionAlt(k))

            fh.write("         Category context: %s\n" % self.getCategoryContextList(k))

            ctL = self.getCategoryExampleList(k)
            if len(ctL) > 0:
                fh.write("    Category example list length %d\n" % len(ctL))
                for ct1, ct2 in ctL:
                    fh.write("      Example   case: %s\n" % ct1)
                    fh.write("      Example detail: %s\n" % ct2)

            ctL = self.getCategoryExampleListAlt(k)
            if len(ctL) > 0:
                fh.write("    Alt category example list length %d\n" % len(ctL))
                for ct1, ct2 in ctL:
                    fh.write("     Alt example   case: %s\n" % ct1)
                    fh.write("     Alt example detail: %s\n" % ct2)

            for v in sorted(uvL):
                fh.write("  Attribute: %s\n" % v)
                fh.write("     Description: %s\n" % self.getDescription(k, v))
                fh.write(" Alt description: %s\n" % self.getDescriptionAlt(k, v))
                fh.write("            Type: %s\n" % self.getTypeCode(k, v))
                fh.write("        Alt Type: %s\n" % self.getTypeCodeAlt(k, v))
                fh.write("  Primitive type: %s\n" % self.getTypePrimitive(k, v))
                fh.write("      Regex type: %s\n" % self.getTypeRegex(k, v))
                fh.write("         Context: %s\n" % self.getContextList(k, v))

                itL = self.getEnumList(k, v)
                if len(itL) > 0:
                    fh.write("      Enum list length %d\n" % len(itL))
                    for it in itL:
                        fh.write("      Enum: %s\n" % it)

                itL = self.getParentList(k, v)
                if len(itL) > 0:
                    fh.write("    Parent list length %d\n" % len(itL))
                    for it in itL:
                        fh.write("      Parent: %s\n" % it)
                itL = self.getChildList(k, v)
                if len(itL) > 0:
                    fh.write("    Child list length %d\n" % len(itL))
                    for it in itL:
                        fh.write("      Child: %s\n" % it)

                itL = self.getExampleList(k, v)
                if len(itL) > 0:
                    fh.write("    Example list length %d\n" % len(itL))
                    for it1, it2 in itL:
                        fh.write("      Example   case: %s\n" % it1)
                        fh.write("      Example detail: %s\n" % it2)

                itL = self.getBoundaryList(k, v)
                if len(itL) > 0:
                    fh.write("    Boundary list length %d\n" % len(itL))
                    for (it1, it2) in itL:
                        fh.write("      Boundary condition (min,max):  (%s,%s)\n" % (it1, it2))

                itL = self.getEnumListAlt(k, v)
                if len(itL) > 0:
                    fh.write("      Alt enum list length %d\n" % len(itL))
                    for it in itL:
                        fh.write("      Alt enum: %s\n" % it)

                itL = self.getExampleListAlt(k, v)
                if len(itL) > 0:
                    fh.write("    Alt example list length %d\n" % len(itL))
                    for it1, it2 in itL:
                        fh.write("      Alt example   case: %s\n" % it1)
                        fh.write("      Alt example detail: %s\n" % it2)

                itL = self.getBoundaryListAlt(k, v)
                if len(itL) > 0:
                    fh.write("    Alt boundary list length %d\n" % len(itL))
                    for (it1, it2) in itL:
                        fh.write("      Alt boundary condition (min,max):  (%s,%s)\n" % (it1, it2))

    #
    def __get(self, enumCode, category, attribute=None):
        """ Returns the last occurrence of the input dictionary metadata (enumCode) for the input category/attribute
            encountered in the list of objects stored at the indicated definition index.

        """
        eS = None
        if enumCode not in self.enumD:
            return eS

        if (attribute is not None):
            nm = "_" + category + "." + attribute
        else:
            nm = category

        if nm in self.definitionIndex:
            dObjL = self.definitionIndex[nm]
            for dObj in dObjL:
                dc = dObj.getObj(self.enumD[enumCode][0])
                if dc is not None:
                    atN = self.enumD[enumCode][1]
                    rL = dc.getRowList()
                    if len(rL) > 0:
                        row = rL[0]
                        if atN is not None:
                            if (dc.hasAttribute(atN)):
                                eS = row[dc.getIndex(atN)]
                        else:
                            tL = []
                            for rv in row:
                                tL.append(rv)
                            eS = tL
        return eS

    def __getList(self, enumCode, category, attribute=None):
        ''' Return the unique list of values '''
        eL = []
        if enumCode not in self.enumD:
            return eL

        if (attribute is not None):
            nm = "_" + category + "." + attribute
        else:
            nm = category

        if nm in self.definitionIndex:
            dObjL = self.definitionIndex[nm]
            for dObj in dObjL:
                dc = dObj.getObj(self.enumD[enumCode][0])
                if dc is not None:
                    atN = self.enumD[enumCode][1]
                    for row in dc.getRowList():
                        if atN is not None:
                            if (dc.hasAttribute(atN)):
                                eL.append(row[dc.getIndex(atN)])
                        else:
                            tL = []
                            for rv in row:
                                tL.append(rv)
                            eL.append(tL)
        eS = set(eL)
        eL = list(eS)
        return eL

    def __getListAll(self, enumCode, category, attribute=None):
        ''' Return a list of all values  '''
        eL = []
        if enumCode not in self.enumD:
            return eL

        if (attribute is not None):
            nm = "_" + category + "." + attribute
        else:
            nm = category

        if nm in self.definitionIndex:
            dObjL = self.definitionIndex[nm]
            for dObj in dObjL:
                dc = dObj.getObj(self.enumD[enumCode][0])
                if dc is not None:
                    atN = self.enumD[enumCode][1]
                    for row in dc.getRowList():
                        if atN is not None:
                            if (dc.hasAttribute(atN)):
                                eL.append(row[dc.getIndex(atN)])
                        else:
                            tL = []
                            for rv in row:
                                tL.append(rv)
                            eL.append(tL)

        return eL

    def getMethodIndex(self):
        return self.methodIndex

    def categoryPart(self, name):
        tname = ""
        if name.startswith("_"):
            tname = name[1:]
        else:
            tname = name

        i = tname.find(".")
        if i == -1:
            return tname
        else:
            return tname[:i]

    def attributePart(self, name):
        i = name.find(".")
        if i == -1:
            return None
        else:
            return name[i + 1:]

    def __getIndex(self):
        for d in self.dictObjList:
            name = d.getName()
            type = d.getType()
            #
            if name not in self.fullIndex:
                self.fullIndex[name] = []
            self.fullIndex[name].append(d)
            #
            if (type == "definition" and d.isCategory()):
                if name not in self.catNameIndex:
                    self.catNameIndex[name] = []
                if name not in self.definitionIndex:
                    self.definitionIndex[name] = []
                self.definitionIndex[name].append(d)

            elif (type == "definition" and d.isAttribute()):
                catN = CifName.categoryPart(name)
                attN = CifName.attributePart(name)
                if catN not in self.catNameIndex:
                    self.catNameIndex[catN] = []
                if attN not in self.catNameIndex:
                    self.catNameIndex[catN].append(attN)
                if name not in self.definitionIndex:
                    self.definitionIndex[name] = []
                self.definitionIndex[name].append(d)
            elif (type == "data"):
                for nm in d.getObjNameList():
                    if nm not in self.dataIndex:
                        self.dataIndex[nm] = d.getObj(nm)
            else:
                pass

    def getDefinitionIndex(self):
        return self.definitionIndex

    def getFullIndex(self):
        return self.fullIndex

    def getMethod(self, id):
        if id in self.methodDict:
            return self.methodDict[id]
        else:
            return None

    def getCategoryList(self):
        return list(self.catNameIndex.keys())

    def getCategoryIndex(self):
        return self.catNameIndex

    def __getMethods(self):
        self.methodDict = {}
        self.methodIndex = {}
        for ob in self.dictObjList:
            if (ob.getType() == 'data'):
                ml = ob.getObj('method_list')
                if ml is not None:
                    for row in ml.getRowList():
                        if ml.hasAttribute('id') and ml.hasAttribute('inline'):
                            mth = MethodDefinition(row[ml.getIndex('id')], row[ml.getIndex('code')], row[ml.getIndex('language')], row[ml.getIndex('inline')])
                            self.methodDict[row[ml.getIndex('id')]] = mth

                ml = ob.getObj('datablock_methods')
                if ml is not None:
                    for row in ml.getRowList():
                        if ml.hasAttribute('method_id'):
                            mth = MethodReference(row[ml.getIndex('method_id')], 'datablock', ob.getName(), None)
                            if (ob.getName() in self.methodIndex):
                                self.methodIndex[ob.getName()].append(mth)
                            else:
                                self.methodIndex[ob.getName()] = []
                                self.methodIndex[ob.getName()].append(mth)
            elif (ob.getType() == 'definition'):
                mi = ob.getObj('category_methods')
                if mi is not None:
                    for row in mi.getRowList():
                        if mi.hasAttribute('method_id'):
                            mth = MethodReference(row[mi.getIndex('method_id')], 'category', ob.getName(), None)
                            if (ob.getName() in self.methodIndex):
                                self.methodIndex[ob.getName()].append(mth)
                            else:
                                self.methodIndex[ob.getName()] = []
                                self.methodIndex[ob.getName()].append(mth)
                mi = ob.getObj('item_methods')
                if mi is not None:
                    for row in mi.getRowList():
                        if mi.hasAttribute('method_id'):
                            mth = MethodReference(row[mi.getIndex('method_id')], 'attribute',
                                                  CifName.categoryPart(ob.getName()),
                                                  CifName.attributePart(ob.getName()))
                            if (ob.getName() in self.methodIndex):
                                self.methodIndex[ob.getName()].append(mth)
                            else:
                                self.methodIndex[ob.getName()] = []
                                self.methodIndex[ob.getName()].append(mth)
            else:
                pass
        return self.methodIndex

    def __getTypes(self):
        for ob in self.dictObjList:
            if (ob.getType() == 'data'):
                tl = ob.getObj('item_type_list')
                if tl is not None:
                    for row in tl.getRowList():
                        if tl.hasAttribute('code') and tl.hasAttribute('primitive_code') and tl.hasAttribute('construct'):
                            self.typesDict[row[tl.getIndex('code')]] = (row[tl.getIndex('primitive_code')], row[tl.getIndex('construct')])

    def dumpCategoryIndex(self, fh=sys.stdout):
        for k, vL in self.catNameIndex.items():
            uvL = list(set(vL))
            fh.write("Category: %s has %d attributes\n" % (k, len(uvL)))
            for v in sorted(uvL):
                fh.write("  Attribute: %s\n" % v)

    def dumpMethods(self, fh=sys.stdout):
        for k, vL in self.methodIndex.items():
            fh.write("Method index key: %s length %d\n" % (k, len(vL)))
            for v in vL:
                v.printIt(fh)
        #
        for k, vL in self.methodIndex.items():
            fh.write("\n------------------------------------\n")
            fh.write("Method inline text: %s\n" % k)
            for v in vL:
                fh.write("Method text: %s\n" % self.getMethod(v.getId()).getInline())


class PdbxBuild(object):

    def __init__(self, verbose=False, log=sys.stderr):
        #
        self.verbose = verbose
        self.__lfh = log
        #
        self.dictFileName = None
        self.dataFileName = None
        # list of dictionary data & definition containers
        self.myDictList = []
        self.myDictIndex = {}
        self.dApi = None
        #
        # list of data containers
        self.myDataList = []
        self.myDataIndex = {}
        #
        # list of ddl dictionary data and definition containers -
        self.ddlFileName = None
        self.myDdlDictList = []
        self.myDdlDictIndex = {}
        self.ddlParentIndex = {}
        self.ddlChildIndex = {}
        self.ddlImplicitList = []
        self.ddlApi = None
        #
        self.renamedDictList = []
        self.renamedDictIndex = {}
        #

    def setDdlFilePath(self, fileName):
        self.ddlFileName = fileName
        self.__readDdlDictionary()

    def __readDdlDictionary(self):
        ''' Read the DDL dictionary and populate some key internal data
        structures.
        '''
        ifh = open(self.ddlFileName, "r")
        pRd = PdbxReader(ifh)
        pRd.read(self.myDdlDictList)
        ifh.close()
        #        i=0
        #        for d in self.myDdlDictList:
        #            self.myDdlDictIndex[d.getName()]=i
        #            i+=1
        self.ddlApi = DictionaryApi(self.myDdlDictList, True)
        self.myDdlDictIndex = self.ddlApi.getFullIndex()
        for nm, dObjL in self.myDdlDictIndex.items():
            for dObj in dObjL:
                if dObj.getType() == "definition" and dObj.isAttribute():
                    catName = CifName.categoryPart(nm)
                    attName = CifName.attributePart(nm)
                    self.ddlParentIndex[nm] = self.ddlApi.getParentList(catName, attName)
                    self.ddlChildIndex[nm] = self.ddlApi.getChildList(catName, attName)
        #
        self.ddlImplicitList = self.ddlApi.getImplicitList()

        return self.myDdlDictList

    def __addItemAlias(self, dObj, oldName, newName):
        if "item_aliases" in dObj.getObjNameList():
            aCat = dObj.getObj("item_aliases")
        else:
            aCat = DataCategory("item_aliases")
            aCat.appendAttribute("alias_name")
            aCat.appendAttribute("dictionary")
            aCat.appendAttribute("version")
            aCat.append({})
            dObj.append(aCat)

        aCat.setValue("cif_rcsb.dic", "dictionary", 1)
        aCat.setValue("1.1", "version", 1)
        aCat.setValue(oldName, "alias_name", 1)

    def __addCategoryDefinitionContext(self, dObj, type):

        if "pdbx_category_context" in dObj.getObjNameList():
            aCat = dObj.getObj("pdbx_category_context")
        else:
            aCat = DataCategory("pdbx_category_context")
            aCat.appendAttribute("type")
            aCat.appendAttribute("category_id")
            aCat.append({})
            dObj.append(aCat)

        aCat.setValue(type, "type", 1)
        aCat.setValue(dObj.getName(), "category_id", 1)

    def __addItemDefinitionContext(self, dObj, type):

        if "pdbx_item_context" in dObj.getObjNameList():
            aCat = dObj.getObj("pdbx_item_context")
        else:
            aCat = DataCategory("pdbx_item_context")
            aCat.appendAttribute("type")
            aCat.appendAttribute("item_name")
            aCat.append({})
            dObj.append(aCat)

        aCat.setValue(type, "type", 1)
        aCat.setValue(dObj.getName(), "item_name", 1)

    def __reformatDefinition(self, dObj):
        dataCatList = dObj.getObjNameList()
        for dataCat in dataCatList:
            catObj = dObj.getObj(dataCat)
            catName = catObj.getName()
            attNameList = catObj.getAttributeList()
            for attName in attNameList:
                targetItem = "_" + catName + "." + attName
                if (targetItem in ['_category.description', '_item_description.description']):
                    desT = catObj.getValue(attName, 1)
                    if desT.find("\n") == -1:
                        catObj.setValue(desT.strip(), attName, 1)

    def __rename(self, oldName, newName, nameType, fh=sys.stdout):
        """First perform the rename operation on any definition
           with name 'oldName'.

           """
        if oldName in self.myDictIndex:
            for dObj in self.myDictIndex[oldName]:
                self.__renameDefinition(dObj, oldName, newName, nameType)
                #
                dObj.setName(newName)
                self.renamedDictList.append(dObj)
                # if not self.renamedDictIndex.has_key(newName):
                #    self.renamedDictIndex[newName]=[]
                # self.renamedDictIndex[newName].append(dObj)
        else:
            pass

    def __dumpDefinition(self, dObj, fh):
        name = dObj.getName()
        if dObj.isCategory():
            fh.write("\nCategory definition: %s\n" % name)
        elif dObj.isAttribute():
            fh.write("\nAttribute definition: %s\n" % name)
        else:
            pass

        # Dump the content of the defintion -
        dataCatList = dObj.getObjNameList()
        for dataCat in dataCatList:
            catObj = dObj.getObj(dataCat)
            catName = catObj.getName()
            attNameList = catObj.getAttributeList()
            for attName in attNameList:
                targetItem = "_" + catName + "." + attName
                fh.write("  Category %s attribute: %s\n" % (catName, attName))
                fh.write("  First value: %s\n" % catObj.getValue(attName, 1))

    def __renameDefinition(self, dObj, oldName, newName, nameType):
        #
        # Rename values of category identifiers or item names within the
        # the input definition from oldName to newName.
        #
        eList = ['_category.description', '_category_examples.case', '_item_examples.case', '_item_description.description',
                 '_ndb_category_description,description', '_ndb_category_examples.case',
                 '_ndb_item_examples.case', '_ndb_item_description.description']
        #
        defCatList = dObj.getObjNameList()
        for defCat in defCatList:
            catObj = dObj.getObj(defCat)
            catName = catObj.getName()
            attNameList = catObj.getAttributeList()
            for attName in attNameList:
                curItem = "_" + catName + "." + attName
                if (nameType == "category" and (curItem == "_category.id" or curItem in self.ddlChildIndex['_category.id'])):
                    # rename the category identifier any of its children
                    catObj.replaceValue(oldName, newName, attName)
                elif (nameType == "attribute" and (curItem == "_item.name" or curItem in self.ddlChildIndex['_item.name'])):
                    # rename the attribute identifier any of its children
                    catObj.replaceValue(oldName, newName, attName)
                else:
                    pass
                # Try to catch names in examples and descriptions -
                if (nameType == 'category' and (curItem in eList)):
                    catObj.replaceSubstring(oldName, newName, attName)
                    oldNameU = oldName.upper()
                    newNameU = newName.upper()
                    catObj.replaceSubstring(oldNameU, newNameU, attName)
                elif (nameType == 'attribute' and (curItem in eList)):
                    catObj.replaceSubstring(oldName, newName, attName)
                else:
                    pass

    def getDictionary(self):
        return self.dApi

    def __consolidateDefinitions(self):
        """ Consolidate definitions into a single saveframe section per definition.
        """
        # preserve the original order of sections -
        nList = []
        for dObj in self.myDictList:
            nm = dObj.getName()
            if nm not in nList:
                nList.append(nm)
        #
        for name, dObjL in self.myDictIndex.items():
            if len(dObjL) > 1:
                for d in dObjL[1:]:
                    xList = d.getObjNameList()
                    for nm in xList:
                        if nm not in dObjL[0].getObjNameList():
                            catObj = d.getObj(nm)
                            dObjL[0].append(catObj)

        # create a new list of consolidated objects in original list order
        dList = []
        for nm in nList:
            if nm in self.myDictIndex:
                dl = self.myDictIndex[nm]
                dList.append(dl[0])
            else:
                print("unknown", nm)

        # update lists
        self.myDictList = dList
        self.dApi = DictionaryApi(self.myDictList, True)
        self.myDictIndex = self.dApi.getFullIndex()
        return self.myDictList

    def readDictionary(self, fileName, consolidate=False):
        self.dictFileName = fileName
        ifh = open(fileName, "r")
        pRd = PdbxReader(ifh)
        pRd.read(self.myDictList)
        ifh.close()

        # i=0
        # for d in self.myDictList:
        #    self.myDictIndex[d.getName()]=i
        #    i+=1

        self.dApi = DictionaryApi(self.myDictList, True)
        self.myDictIndex = self.dApi.getFullIndex()
        if (consolidate):
            self.__consolidateDefinitions()

        return self.myDictList

    def getMethods(self):
        return self.dApi.getMethods()

    def getMethod(self, id):
        return self.dApi.getMethod(id)

    def writeDictionary(self, fileName):
        #
        ofh = open(fileName, "w")
        pWr = PdbxWriter(ofh)
        pWr.write(self.myDictList)
        ofh.close()

    def writeRenamedDictionary(self, fileName):
        #
        ofh = open(fileName, "w")
        pWr = PdbxWriter(ofh)
        pWr.write(self.renamedDictList)
        ofh.close()

    def writeDictionaryObjList(self, fileName, objList):
        #
        ofh = open(fileName, "w")
        pWr = PdbxWriter(ofh)
        pWr.write(objList)
        ofh.close()

    def readDataFile(self, fileName):
        self.dataFileName = fileName
        ifh = open(fileName, "r")
        pRd = PdbxReader(ifh)
        pRd.read(self.myDataList)
        ifh.close()
        i = 0
        for d in self.myDataList:
            self.myDataIndex[d.getName()] = i
            i += 1
        return self.myDataList

    def writeDataFile(self, fileName):
        ofh = open(fileName, "w")
        pWr = PdbxWriter(ofh)
        pWr.write(self.myDataList)
        ofh.close()

    def invokeMethods(self, fh=sys.stderr):
        mI = self.dApi.getMethodIndex()
        lenD = len(mI)
        i = 0
        for k, mRefL in mI.items():
            for mRef in mRefL:
                i += 1
                id = mRef.getId()
                type = mRef.getType()
                categoryName = mRef.getCategoryName()
                attributeName = mRef.getAttributeName()
                if (self.verbose):
                    fh.write("\n")
                    fh.write("++++++++++++++++++--------------------\n")
                    fh.write("Invoking dictionary method on file object: %s (%d/%d)\n" % (k, i, lenD))
                    fh.write(" + Method id: %s\n" % id)
                    fh.write(" + Type:      %s\n" % type)
                    fh.write(" + Category:  %s\n" % categoryName)
                    fh.write(" + Attribute: %s\n" % attributeName)
                #
                if type == "datablock":
                    if (self.verbose):
                        fh.write("Invoke datablock method %s\n" % id)
                    # self.invokeDataBlockMethod(type,self.dApi.getMethod(id))
                    # continue
                #
                for db in self.myDataList:
                    if type == "category":
                        if not db.exists(categoryName):
                            dc = DataCategory(categoryName)
                            db.append(dc)
                        dObj = db.getObj(categoryName)
                        dObj.invokeCategoryMethod(type, self.dApi.getMethod(id), db)
                    elif type == "attribute":
                        if not db.exists(categoryName):
                            dc = DataCategory(categoryName)
                            db.append(dc)
                        dObj = db.getObj(categoryName)
                        dObj.invokeAttributeMethod(attributeName, type, self.dApi.getMethod(id), db)
                    elif type == "datablock":
                        fh.write("Invoke datablock method %s\n" % id)
                        db.invokeDataBlockMethod(type, self.dApi.getMethod(id), db)
                    else:
                        pass

    def renameDefinition(self, currentName, NewName, nameType, fh=sys.stdout):
        self.__rename(currentName, NewName, nameType, fh)

    def renameDefinitionList(self, renameList, aliasDict, fh=sys.stdout):
        """Make one pass over all of the containers in the dictionary
           making all of the replacements in the renameList

           """
        for nameType, oldName, newName in renameList:
            if oldName in self.myDictIndex:
                for dObj in self.myDictIndex[oldName]:
                    self.__renameDefinition(dObj, oldName, newName, nameType)
                    dObj.setName(newName)
                    self.renamedDictList.append(dObj)
                    # if not self.renamedDictIndex.has_key(newName):
                    #    self.renamedDictIndex[newName]=[]
                    # self.renamedDictIndex[newName].append(dObj)
                    #
                    if dObj.isAttribute():
                        self.__addItemAlias(dObj, oldName, newName)
                        self.__addItemDefinitionContext(dObj, 'RCSB_LOCAL')
                    elif dObj.isCategory():
                        self.__addCategoryDefinitionContext(dObj, 'RCSB_LOCAL')
            else:
                pass

        # Now re-check for other substitutions in rename list -
        for dObj in self.renamedDictList:
            for nameType, oldName, newName in renameList:
                self.__renameDefinition(dObj, oldName, newName, nameType)
            for oldName, newName in aliasDict.items():
                if oldName.startswith("_"):
                    nameType = "attribute"
                else:
                    nameType = "category"
                self.__renameDefinition(dObj, oldName, newName, nameType)

            self.__reformatDefinition(dObj)

        return self.renamedDictList

    def dumpMethods(self, fh=sys.stdout):
        self.dApi.dumpMethods(fh)

    def dumpDictionary(self, fh=sys.stdout):
        lenD = len(self.myDictList)
        fh.write("\n--------------------------------------------\n")
        fh.write("\n-----------DUMP DICTIONARY------------------\n")
        fh.write("Dictionary object list length is: %d\n" % lenD)
        i = 1
        for dObj in self.myDictList:
            if len(dObj.getName()) > 0:
                fh.write("\n")
                fh.write("++++++++++++++++++--------------------\n")
                fh.write("Dumping dictionary object named: %s (%d/%d)\n" %
                         (dObj.getName(), i, lenD))
                dObj.printIt(fh)
            i += 1
    #

    def dumpDataFile(self, fh=sys.stdout):
        lenD = len(self.myDataList)
        fh.write("\n--------------------------------------------\n")
        fh.write("\n-----------DUMP DATAFILE--------------------\n")
        fh.write("Data object list length is: %d\n" % lenD)
        i = 1
        for dObj in self.myDataList:
            fh.write("\n")
            fh.write("++++++++++++++++++--------------------\n")
            fh.write("Dumping data file object named: %s (%d/%d)\n" %
                     (dObj.getName(), i, lenD))
            dObj.printIt(fh)
            i += 1
    #

    def dumpDictionaryAliases(self, fh=sys.stdout):
        """ Return alias dictionary and a complete list of all definitions (name,type)
        """
        lenD = len(self.myDictList)
        fh.write("\n--------------------------------------------\n")
        fh.write("\n-----------DUMP DICTIONARY------------------\n")
        fh.write("Dictionary object list length is: %d\n" % lenD)
        i = 1
        nList = []
        aDict = {}
        for dObj in self.myDictList:
            type = dObj.getType()
            if (type == "definition"):
                name = str(dObj.getName()).strip()
                if dObj.isCategory():
                    fh.write("  Category: %s\n" % name)
                    nList.append((name, "category"))
                elif dObj.isAttribute():
                    nList.append((name, "attribute"))
                    if name.startswith("_pdbx_") or (name.find(".pdbx_") > 0):
                        fh.write("  PDBX Attribute: %s\n" % name)
                        aliasCat = dObj.getObj("item_aliases")
                        if (aliasCat is not None):
                            # aliasCat.printIt(fh)
                            for row in aliasCat.getRowList():
                                if (aliasCat.hasAttribute("dictionary") and aliasCat.hasAttribute("alias_name")
                                        and (row[aliasCat.getIndex('dictionary')] == "cif_rcsb.dic")):
                                    fh.write("       RCSB alias name is %s\n" % row['alias_name'])
                                    alias = row[aliasCat.getIndex('alias_name')]
                                    aDict[alias] = name
                    elif name.startswith("_ndb_"):
                        fh.write("  NDB Attribute: %s\n" % name)
                    elif name.startswith("_em_"):
                        fh.write("  EM Attribute: %s\n" % name)
                    else:
                        pass
                else:
                    pass
            i += 1
        #
        for k, v in aDict.items():
            fh.write("alias %s  name %s\n" % (k, v))
        #
        return (aDict, nList)



##
# File: IoAdapterCore.py
# Date: 20-Jan-2013  John Westbrook
#
# Updates:
# 20-Jan-2013 jdw Type convert to string or '?'
#  7-May-2014 jdw Make output line length 2048
# 30-Jul-2014 jdw Expose maximum line length as an optional writeFile() parameter.
#  5-Feb-2015 jdw disable maximum line length parameter -
#  6-Dec-2015 jdw add detailed timers
#  6-Dec-2015 jdw add additional filters for category write order -
# 28-Jul-2016 rps readFile(), __readData() methods updated to accept optional "logtag" parameter
# 15-Feb-2017 ep  Correct variable name in exception in __readDataSelect()
#  8-Jan-2018 jdw adapt to new pybind11 bindings -  change logging framework -- handle py2->3
# 10-Jan-2018 jdw complete rewrite for new mmciflib framework.
#
##
"""
Adapter between Python mmCIF API and Pybind11 wrappers for the PDB C++ Core mmCIF Library.

"""
from __future__ import absolute_import
from six.moves import range

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import sys
import time
import os
from future.utils import raise_from

import logging
logger = logging.getLogger(__name__)

HERE = os.path.abspath(os.path.dirname(__file__))


from mmcif.io.IoAdapterBase import IoAdapterBase

from mmcif.api.PdbxContainers import *
from mmcif.api.DataCategory import DataCategory
from mmcif.io.PdbxExceptions import PdbxError, SyntaxError

try:
    from mmcif.core.mmciflib import ParseCifSimple, CifFile, ParseCifSelective, type, CifFileReadDef
except Exception as e:
    sys.path.insert(0, os.path.dirname(os.path.dirname(HERE)))
    from build.lib.mmciflib import ParseCifSimple, CifFile, ParseCifSelective, type, CifFileReadDef


class IoAdapterCore(IoAdapterBase):
    """ Adapter between Python mmCIF API and Pybind11 wrappers for the PDB C++ Core mmCIF Library.
    """

    def __init__(self, *args, **kwargs):
        super(IoAdapterCore, self).__init__(*args, **kwargs)

    def readFile(self, inputFilePath, enforceAscii=True, selectList=None, excludeFlag=False, logFilePath=None, outDirPath=None, cleanUp=True, **kwargs):
        """Parse the data blocks in the input mmCIF format data file into list of DataContainers().  The data category content within each data block
           is stored a collection of DataCategory objects within each DataContainer.

        Args:
            inputFilePath (string): Input file path
            enforceAscii (bool, optional): Flag to requiring pre-filtering operation to convert input file to ASCII encoding. See encoding error options.
            selectList (List, optional):  List of data category names to be extracted or excluded from the input file (default: select/extract)
            excludeFlag (bool, optional): Flag to indicate selectList should be treated as an exclusion list
            logFilePath (string, optional): Log file path (if not provided this will be derived from the input file.)
            outDirPath (string, optional): Path for translated/reencoded files and default logfiles.
            cleanUp (bool, optional): Flag to automatically remove logs and temporary files on exit.
            **kwargs: Placeholder for missing keyword arguments.

        Returns:
            List of DataContainers: Contents of input file parsed into a list of DataContainer objects.
        """
        if len(kwargs):
            logger.warn("Unsupported keyword arguments %s" % kwargs.keys())
        asciiFilePath = None
        filePath = str(inputFilePath)
        try:
            #
            lPath = logFilePath
            if not lPath:
                lPath = self._getDefaultFileName(filePath, fileType='cif-parser-log', outDirPath=outDirPath)
            #
            self._setLogFilePath(lPath)
            #
            if not self._fileExists(filePath):
                return []
            #
            tPath = filePath
            if enforceAscii:
                asciiFilePath = self._getDefaultFileName(filePath, fileType='cif-parser-ascii', fileExt='cif', outDirPath=outDirPath)
                encodingErrors = 'xmlcharrefreplace' if self._useCharRefs else 'ignore'
                logger.debug("Filtering input file to %s using encoding errors as %s" % (asciiFilePath, encodingErrors))
                ok = self._toAscii(filePath, asciiFilePath, chunkSize=5000, encodingErrors=encodingErrors)
                if ok:
                    tPath = asciiFilePath
            #
            readDef = None
            if selectList and len(selectList) > 0:
                readDef = self.__getSelectionDef(selectList, excludeFlag)
            #
            containerL, diagL = self.__readData(tPath, readDef=readDef, cleanUp=cleanUp, logFilePath=lPath, maxLineLength=self._maxInputLineLength)
            #
            self._cleanupFile(asciiFilePath and cleanUp, asciiFilePath)
            #
            return containerL
        except (PdbxError, SyntaxError) as ex:
            self._cleanupFile(asciiFilePath and cleanUp, asciiFilePath)
            if self._raiseExceptions:
                raise_from(ex, None)
                # raise ex from None
        except Exception as e:
            self._cleanupFile(asciiFilePath and cleanUp, asciiFilePath)
            msg = "Failing read for %s with %s" % (filePath, str(e))
            self._logError(msg)

        return []

    def getReadDiags(self):
        """Recover the diagnostics for the previous readFile() operation.readFile

           Returns:
             list of strings:  List of parsing and processing diagnostics from last readFile() operation
        """
        return self._readLogRecords()

    def __getSelectionDef(self, selectList, excludeFlag):
        """ Internal method to package selection/exclusion list for the C++ parser library.

            Returns:
               CifFileReadDef() object:  object prepared for parsing library
        """
        try:
            readDef = CifFileReadDef()
            if excludeFlag:
                readDef.SetCategoryList(selectList, type.D)
            else:
                readDef.SetCategoryList(selectList, type.A)
            return readDef
        except Exception as e:
            msg = "Failing read selection with %s" % str(e)
            self._logError(msg)
        return None

    def __processReadLogFile(self, inputFilePath):
        """ Internal method to process logfiles and either log errors or raise exceptions (See: Class PdbxExceptions).
            The behavior is controlled by the class attribute _raiseExcetions.

            Returns:
             list of strings:  List of records in the input log file
        """
        diagL = self._readLogRecords()
        #
        if diagL:
            numErrors = 0
            numSyntaxErrors = 0
            numWarnings = 0
            for diag in diagL:
                if 'ERROR' in diag:
                    numErrors += 1
                if 'WARN' in diag:
                    numWarnings += 1
                if 'syntax' in diag.lower():
                    numSyntaxErrors += 1
            #
            logger.debug("%s syntax errors %d  warnings %d all errors %d" % (inputFilePath, numSyntaxErrors, numWarnings, numErrors))
            #
            if numSyntaxErrors and self._raiseExceptions:
                raise SyntaxError("%s syntax errors %d  all errors %d" % (inputFilePath, numSyntaxErrors, numErrors))
            elif numErrors and self._raiseExceptions:
                raise PdbxError("%s error count is %d" % (inputFilePath, numErrors))
            elif numErrors:
                logger.error("%s syntax errors %d  all errors %d" % (inputFilePath, numSyntaxErrors, numErrors))
            if numWarnings:
                logger.warn("%s warnings %d" % (pdbxFilePath, numWarnings))

        return diagL

    def __processContent(self, cifFileObj):
        """Internal method to transfer parsed data from the wrapped input C++ CifFile object into
        the list of Python DataContainer objects.

        Args:
            cifFileObj (wrapped CifFile object): Wrapped input C++ CifFile object

        Returns:
            list of DataContainer objects:   List of Python DataContainer objects

        """
        containerList = []
        containerNameList = []
        try:
            # ----- Repackage the data content  ----
            #
            containerList = []
            containerNameList = []
            containerNameList = list(cifFileObj.GetBlockNames(containerNameList))
            for containerName in containerNameList:
                #
                aContainer = DataContainer(containerName)
                #
                block = cifFileObj.GetBlock(containerName)
                tableNameList = []
                tableNameList = list(block.GetTableNames(tableNameList))

                for tableName in tableNameList:
                    table = block.GetTable(tableName)
                    attributeNameList = list(table.GetColumnNames())
                    numRows = table.GetNumRows()
                    rowList = []
                    for iRow in range(0, numRows):
                        row = table.GetRow(iRow)
                        # row = table.GetRow(iRow).decode('unicode_escape').encode('utf-8')
                        # row = [p.encode('ascii', 'xmlcharrefreplace') for p in table.GetRow(iRow)]
                        rowList.append(list(row))
                    aCategory = DataCategory(tableName, attributeNameList, rowList, copyInputData=False)
                    aContainer.append(aCategory)
                containerList.append(aContainer)
        except Exception as e:
            msg = "Failing packaging with %s" % str(e)
            self._logError(msg)

        return containerList

    def __readData(self, inputFilePath, readDef=None, maxLineLength=1024, logFilePath=None, cleanUp=False):
        """Internal method to read input file and return data as a list of DataContainer objects.
        readDef optionally contains a selection of data categories to be returned.    Diagnostics
        will be written to logFilePath (persisted if cleanuUp=False).

        Args:
            inputFilePath (string):  input file path
            readDef (CifFileReadDef object, optional): wrapped CifFileReadDef() object
            maxLineLength (int, optional): Maximum supported line length on input
            logFilePath (string, optional): Log file path
            cleanUp (bool, optional): Flag to remove temporary files on exit

        Returns:
            Tuple of lists : DataContainer List, Diagnostics (string) List

        """
        #
        startTime = time.clock()
        containerList = []
        diagL = []
        try:
            if readDef:
                cifFileObj = ParseCifSelective(
                    inputFilePath,
                    readDef,
                    verbose=self._verbose,
                    intCaseSense=0,
                    maxLineLength=maxLineLength,
                    nullValue="?",
                    parseLogFileName=logFilePath)
            else:
                cifFileObj = ParseCifSimple(inputFilePath, verbose=self._verbose, intCaseSense=0, maxLineLength=maxLineLength, nullValue="?", parseLogFileName=logFilePath)
            #
            # ---  Process/Handle read errors   ----
            #
            diagL = self.__processReadLogFile(inputFilePath)
            logger.debug("Diagnostic count %d values %r" % (len(diagL), diagL))
            #
            if self._timing:
                stepTime1 = time.clock()
                logger.info("Timing parsed %r in %.4f seconds" % (inputFilePath, stepTime1 - startTime))
            #
            containerList = self.__processContent(cifFileObj)
            #
            self._cleanupFile(cleanUp, logFilePath)
            if self._timing:
                stepTime2 = time.clock()
                logger.info("Timing api load in %.4f seconds read time %.4f seconds\n" %
                            (stepTime2 - stepTime1, stepTime2 - startTime))
            #
            return containerList, diagL
        except (PdbxError, SyntaxError) as ex:
            self._cleanupFile(cleanUp, logFilePath)
            if self._raiseExceptions:
                raise_from(ex, None)
        except Exception as e:
            self._cleanupFile(cleanUp, logFilePath)
            msg = "Failing read for %s with %s" % (inputFilePath, str(e))
            self._logError(msg)

        return containerList, diagL

    def writeFile(self, outputFilePath, containerList=None, maxLineLength=900, enforceAscii=True,
                  lastInOrder=['pdbx_nonpoly_scheme', 'pdbx_poly_seq_scheme', 'atom_site', 'atom_site_anisotrop'], selectOrder=None, **kwargs):
        """Write input list of data containers to the specified output file path in mmCIF format.

        Args:
            outputFilePath (string): output file path
            containerList (list DataContainer objects, optional):
            maxLineLength (int, optional): Maximum length of output line (content is wrapped beyond this length)
            enforceAscii (bool, optional): Filter output (not implemented - content must be ascii compatible on input)
            lastInOrder (list of category names, optional): Move data categories in this list to end of each data block
            selectOrder (list of category names, optional): Write only data categories on this list.
            **kwargs: Placeholder for unsupported key value pairs

        Returns:
            bool: Completion status


        """
        containerL = containerList if containerList else []
        if len(kwargs):
                logger.warn("Unsupported keyword arguments %s" % kwargs.keys())
        try:
            startTime = time.clock()
            logger.debug("write container length %d\n" % len(containerL))
            # (CifFile args: placeholder, verbose: bool, caseSense: Char::eCompareType, maxLineLength: int, nullValue: str)
            cF = CifFile(True, self._verbose, 0, maxLineLength, '?')
            for container in containerL:
                containerName = container.getName()
                logger.debug("writing container %s\n" % containerName)
                cF.AddBlock(containerName)
                block = cF.GetBlock(containerName)
                #
                # objNameList = container.getObjNameList()
                # logger.debug("write category length %d\n" % len(objNameList))
                #
                # Reorder/Filter - container object list-
                objNameList = container.filterObjectNameList(lastInOrder=lastInOrder, selectOrder=selectOrder)
                logger.debug("write category names  %r\n" % objNameList)
                #
                for objName in objNameList:
                    name, attributeNameList, rowList = container.getObj(objName).get()
                    table = block.AddTable(name)
                    for attributeName in attributeNameList:
                        table.AddColumn(attributeName)
                    try:
                        rLen = len(attributeNameList)
                        for ii, row in enumerate(rowList):
                            table.AddRow()
                            table.FillRow(ii, [str(row[jj]) if row[jj] is not None else '?' for jj in range(0, rLen)])
                    except Exception as e:
                        logger.error("Exception for %s preparing data for writing %s" % (outputFilePath, str(e)))
                    #
                    block.WriteTable(table)
            #
            if self._timing:
                stepTime1 = time.clock()
                logger.info("Timing %d container(s) api loaded in %.4f seconds" % (len(containerL), stepTime1 - startTime))
            if (self._debug):
                self.__dumpBlocks(cF)
            cF.Write(str(outputFilePath))
            if self._timing:
                stepTime2 = time.clock()
                logger.info("Timing %d container(s) written in %.4f seconds total time %.4f" %
                            (len(containerList), stepTime2 - stepTime1, stepTime2 - startTime))
            return True

        except Exception as e:
            msg = "Write failing for file %s with %s" % (outputFilePath, str(e))
            self._logError(msg)
        return False

    def __dumpBlocks(self, cf):
        """Internal method to log the contents of the input wrapped CifFile object.

        Args:
            cf (CifFile object): wrapped CifFile object.
        """
        try:
            logger.info("cif file %r" % cf)
            blockNameList = []
            blockNameList = cf.GetBlockNames(blockNameList)
            #
            logger.info(" block name list %r" % repr(blockNameList))
            for blockName in blockNameList:
                #
                block = cf.GetBlock(blockName)
                tableNameList = []
                tableNameList = list(block.GetTableNames(tableNameList))
                logger.info("tables name list %r" % repr(tableNameList))
                for tableName in tableNameList:
                    logger.info("table name %r" % tableName)
                    ok = block.IsTablePresent(tableName)
                    logger.info("table present %r" % ok)
                    table = block.GetTable(tableName)

                    attributeNameList = list(table.GetColumnNames())
                    logger.info("Attribute name list %r" % repr(attributeNameList))
                    numRows = table.GetNumRows()
                    logger.info("row length %r" % numRows)
                    for iRow in range(0, numRows):
                        row = table.GetRow(iRow)
                        logger.info("Attribute name list %r" % row)
        except Exception as e:
            logger.exception("dump failing with %s\n" % str(e))


if __name__ == '__main__':
    pass



##
#
# File:     DataCategoryBase.py
# Original: 02-Feb-2009   jdw
#
# Update:
#   14-Nov-2012   jdw refactoring
#   19-Nov-2012   jdw subclass UserList -
#                     _rowList becomes data
#   27-Nov-2012   jdw revise method removeAttribute()
#    7-Jan-2013   jdw add getAttributeIndexDict()
#   29-Jun-2013   jdw add removeRow()
#   24-Jun-2015   jdw add getRowAttributeDict(self, index) and getRowItemDict(self, index)
#   01-Aug-2017   jdw migrate portions to public repo
##
"""

A collection of container classes supporting the PDBx/mmCIF storage model.

"""
from __future__ import absolute_import
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"


from six.moves import range
from six.moves import zip


from six.moves import UserList
import sys
import copy
import logging
logger = logging.getLogger(__name__)


class DataCategoryBase(UserList):

    """ Base object definition for a data category -

        This class subclasses UserList and implements many list-like features for
        row data managed by this class.

    """

    def __init__(self, name, attributeNameList=None, rowList=None, raiseExceptions=True, copyInputData=True):
        self._name = name
        if copyInputData:
            self._attributeNameList = copy.deepcopy(attributeNameList) if attributeNameList is not None else []
            self.data = copy.deepcopy(rowList) if rowList is not None else []
        else:
            self._attributeNameList = attributeNameList if attributeNameList is not None else []
            self.data = rowList if rowList is not None else []
        self._itemNameList = []
        self.__mappingType = "DATA"
        self._raiseExceptions = raiseExceptions
        self._copyInputData = copyInputData
        #
        super(DataCategoryBase, self).__init__(self.data)
        #
        # Derived class data -
        #
        self._catalog = {}
        self._numAttributes = 0
        #
        self._isPy3 = sys.version_info[0] == 3
        if self._isPy3:
            self._string_types = str
        else:
            self._string_types = basestring
        self.__setup()

    def __setup(self):
        self._numAttributes = len(self._attributeNameList)
        self._catalog = {}
        for attributeName in self._attributeNameList:
            attributeNameLC = attributeName.lower()
            self._catalog[attributeNameLC] = attributeName
    #
    # Setters/appenders
    #

    def setName(self, name):
        self._name = name

    def setRowList(self, rowList):
        if self._copyInputData:
            self.data = copy.deepcopy(rowList)
        else:
            self.data = rowList

    def setAttributeNameList(self, attributeNameList):
        if self._copyInputData:
            self._attributeNameList = copy.deepcopy(attributeNameList)
        else:
            self._attributeNameList = attributeNameList
        self.__setup()

    def appendAttribute(self, attributeName):
        attributeNameLC = attributeName.lower()
        if attributeNameLC in self._catalog:
            i = self._attributeNameList.index(self._catalog[attributeNameLC])
            self._attributeNameList[i] = attributeName
            self._catalog[attributeNameLC] = attributeName
        else:
            self._attributeNameList.append(attributeName)
            self._catalog[attributeNameLC] = attributeName
            #
        self._numAttributes = len(self._attributeNameList)
        return self._numAttributes

    ##
    # Getters
    ##
    def get(self):
        return (self._name, self._attributeNameList, self.data)

    def getName(self):
        return self._name

    def getAttributeList(self):
        return self._attributeNameList

    def getAttributeCount(self):
        return len(self._attributeNameList)

    def getAttributeIndex(self, attributeName):
        try:
            return self._attributeNameList.index(attributeName)
        except Exception as e:
            pass
        return -1

    def getAttributeIndexDict(self):
        rD = {}
        for ii, attributeName in enumerate(self._attributeNameList):
            rD[attributeName] = ii
        return rD

    def getIndex(self, attributeName):
        return self.getAttributeIndex(attributeName)

    def hasAttribute(self, attributeName):
        return attributeName in self._attributeNameList

    def getItemNameList(self):
        return self.__updateItemLabels()

    def getRowList(self):
        return self.data

    def getRowCount(self):
        return (len(self.data))

    def getRow(self, index):
        try:
            return self.data[index]
        except Exception as e:
            if self._raiseExceptions:
                raise e
        return []

    def getRowAttributeDict(self, index):
        rD = {}
        try:
            for ii, v in enumerate(self.data[index]):
                rD[self._attributeNameList[ii]] = v
            return rD
        except Exception as e:
            if self._raiseExceptions:
                raise e

        return rD

    def getRowItemDict(self, index):
        rD = {}
        try:
            self.__updateItemLabels()
            for ii, v in enumerate(self.data[index]):
                rD[self._itemNameList[ii]] = v
            return rD
        except Exception as e:
            if self._raiseExceptions:
                raise e

        return rD

    def getAttributeValueList(self, attributeName):
        """ Return a list of
        """
        try:
            idx = self.getAttributeIndex(attributeName)
            rL = [row[idx] for row in self.data]
            return rL
        except Exception as e:
            raise e

    def removeRow(self, index):
        try:
            del self.data[index]
            return True
        except Exception as e:
            if self._raiseExceptions:
                raise e

        return False

    def removeAttribute(self, attributeName):
        """ Remove the attribute from the attribute list along with any
            corresponding row data.
        """
        idx = self.getAttributeIndex(attributeName)
        if idx != -1:
            try:
                del self._attributeNameList[idx]
                for row in self.data:
                    try:
                        del row[idx]
                    except:
                        pass
                self.__setup()
                return True
            except:
                return False

    ##
    ##
    ##
    def __updateItemLabels(self):
        """  Internal method to create mmCIF style item names for the current attribute
             list.
        """
        self._itemNameList = []
        for atName in self._attributeNameList:
            self._itemNameList.append("_" + str(self._name) + "." + atName)
        #
        return self._itemNameList

    def __alignLabels(self, row):
        """  Internal method which aligns the list of input attributes with row data.

             If there are fewer labels than data elements in a row, then placeholder labels
             are creeated (e.g. "unlabeled_#")

        """
        if len(row) > len(self._attributeNameList):
            for i in range(len(self._attributeNameList), len(row) - 1):
                self._attributeNameList.insert(i, "unlabeled_" + str(i))
            if self.__mappingType == "ITEM":
                self.__updateItemLabels()

    def setMapping(self, mType):
        """  Controls the manner in which this class returns data when accessed by
             index or in the context of an iterator:

             DATA      = list of row data elements as these were input. [default]

             ATTRIBUTE = row returned as a dictionary with attribute key

             ITEM      = row returned as a dictionary with item key

        """
        if mType in ['DATA', 'ATTRIBUTE', 'ITEM']:
            self.__mappingType = mType
            return True
        else:
            return False

    def __str__(self):
        ans = 'name:%r\nattrbuteList: %r\nData: %r\n' % (self._name, self._attributeNameList, list(self.data),)
        return ans

    def __repr__(self):
        return self.__class__.__name__ + '(' + str(self) + ')'

    def __iter__(self):
        for d in self.data:
            yield self.__applyMapping(d)

    def __getitem__(self, idx):
        return self.__applyMapping(self.data[idx])

    def __setitem__(self, idx, value):
        dL = self.__extractMapping(value)
        self.data[idx] = dL

    def __applyMapping(self, d):
        if self.__mappingType == "DATA":
            return d
        elif self.__mappingType == "ATTRIBUTE":
            self.__alignLabels(d)
            return dict(list(zip(self._attributeNameList, d)))
        elif self.__mappingType == "ITEM":
            self.__alignLabels(d)
            self.__updateItemLabels()
            return dict(list(zip(self._itemNameList, d)))

    def __extractMapping(self, d):
        try:
            if self.__mappingType == "DATA":
                return d
            elif self.__mappingType == "ATTRIBUTE":
                rL = []
                for k, v in d.items():
                    rL.insert(self._attributeNameList.index(k), v)
                return rL
            elif self.__mappingType == "ITEM":
                rL = []
                for k, v in d.items():
                    rL.insert(self._itemNameList.index(k), v)
                return rL
        except:
            raise IndexError

    def cmpAttributeNames(self, dcObj):
        """ Compare the attributeNameList in current data category (dca) and input data category .

            Return: (current attributes not in dcObj), (attributes common to both), (attributes in dcObj not in current data category)
        """
        sa = set(self.getAttributeList())
        sb = set(dcObj.getAttributeList())
        return tuple(sa - sb), tuple(sa & sb), tuple(sb - sa)

    def cmpAttributeValues(self, dcObj):
        """ Compare the values by attribute for current data category (dca) and input data category.
            The comparison is performed independently for the values of corresponding attributes.
            Length differences are treated inequality out of hand.

            Return: [(attributeName, values equal flag (bool)), (attributeName, values equal flag (bool), ...]


            Note on slower alternative
            >>> import collections
            >>> compare = lambda x, y: collections.Counter(x) == collections.Counter(y)
        """
        try:
            rL = []
            sa = set(self.getAttributeList())
            sb = set(dcObj.getAttributeList())
            atComList = list(sa & sb)
            #
            lenEq = self.getRowCount() == dcObj.getRowCount()
            for at in atComList:
                if lenEq:
                    same = sorted(self.getAttributeValueList(at)) == sorted(dcObj.getAttributeValueList(at))
                else:
                    same = False
                rL.append((at, same))
            return rL
        except Exception as e:
            raise e

    def __eq__(self, other):
        """Override the default Equals behavior"""
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other):
        """Define a non-equality test"""
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __hash__(self):
        """Override the default hash behavior (that returns the id or the object)"""
        return hash(tuple(sorted(self.__dict__.items())))

    #



##
# File:      Method.py
# Orignal:   Aug 12, 2013   Jdw
#
# Updates:
#   01-Aug-2017   jdw migrate portions to public repo
#
##
"""
Utility classes for applying dictionary methods on PDBx/mmCIF data files.
"""

from __future__ import absolute_import

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"



import sys


class MethodDefinition(object):

    def __init__(self, method_id, code='calculate', language='Python', inline=None):
        self.method_id = method_id
        self.language = language
        self.code = code
        self.inline = inline

    def getId(self):
        return self.method_id

    def getLanguage(self):
        return self.language

    def getInline(self):
        return self.inline

    def printIt(self, fh=sys.stdout):
        fh.write("------------- Method definition -------------\n")
        fh.write("Id:           %s\n" % self.method_id)
        fh.write("Code:         %s\n" % self.code)
        fh.write("Language:     %s\n" % str(self.language))
        fh.write("Inline text:  %s\n" % str(self.inline))


class MethodReference(object):

    def __init__(self, method_id, type='attribute', category=None, attribute=None):
        self.method_id = method_id
        self.type = type
        self.categoryName = category
        self.attributeName = attribute

    def getId(self):
        return self.method_id

    def getType(self):
        return self.type

    def getCategoryName(self):
        return self.categoryName

    def getAttributeName(self):
        return self.attributeName

    def printIt(self, fh=sys.stdout):
        fh.write("--------------- Method Reference -----------------\n")
        fh.write("Id:             %s\n" % self.method_id)
        fh.write("Type:           %s\n" % self.type)
        fh.write("Category name:  %s\n" % str(self.categoryName))
        fh.write("Attribute name: %s\n" % str(self.attributeName))



##
# File:  PdbxReader.py
# Date:  2012-01-09  Jdw  Adapted from PdbxParser
#
# Updates:
#
# 2012-01-09 - (jdw) Separate reader and writer classes.
#
# 2012-09-02 - (jdw)  Revise tokenizer to better handle embedded quoting.
#
# 2013-08-20 - (jdw)  make current container the prior data container when leaving a save frame.
#
# 2013-19-14 - (jdw)  change the treatment of stop_ tokens --
##
"""
Utilities for reading mmCIF for data files and dictionaries.

Acknowledgements:

 The tokenizer used in this module is modeled after the clever parser design
 used in the PyMMLIB package.

 PyMMLib Development Group
 Authors: Ethan Merritt: merritt@u.washington.edu  & Jay Painter: jay.painter@gmail.com
 See:  http://pymmlib.sourceforge.net/

"""
# from __future__ import absolute_import, generator_stop
from __future__ import absolute_import

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"


import re
import sys

import logging
logger = logging.getLogger(__name__)

from mmcif.api.PdbxContainers import *
from mmcif.api.DataCategory import DataCategory
from mmcif.io.PdbxExceptions import PdbxError, SyntaxError


class PdbxReader(object):
    """ Utilities for reading mmCIF for data files and dictionaries.

    """

    def __init__(self, ifh):
        """  ifh - input file handle returned by open()
        """
        #
        self.__curLineNumber = 0
        self.__ifh = ifh
        self.__stateDict = {"data": "ST_DATA_CONTAINER",
                            "loop": "ST_TABLE",
                            "global": "ST_GLOBAL_CONTAINER",
                            "save": "ST_DEFINITION",
                            "stop": "ST_STOP"}

    def read(self, containerList, selectList=None, excludeFlag=False):
        """
        Appends to input list of definition and data containers.

        return

        """
        sL = selectList if selectList else []
        catSelectD = {k: k for k in sL}
        self.__curLineNumber = 0
        try:
            self.__parser(self.__tokenizer(self.__ifh), containerList, categorySelectionD=catSelectD, excludeFlag=excludeFlag)
        except RuntimeError as e:
            # will be raised at the end of token iterator - not an error -
            logger.debug("Normal termination after reading %d lines with %s" % (self.__curLineNumber, str(e)))
        except StopIteration:
            # will be raised at the end of token iterator - not an error -
            logger.debug("Normal termination after reading %d lines" % self.__curLineNumber)
        except SyntaxError as e:
            logger.debug("Caught syntax exception at %d" % self.__curLineNumber)
            raise e
        except UnicodeDecodeError as e:
            logger.debug("Caught character encoding exception at %d" % self.__curLineNumber)
            raise PdbxError("Character encoding error at line %d" % self.__curLineNumber)
        except Exception as e:
            raise PdbxError("Failing at line %d with %s" % (self.__curLineNumber, str(e)))
        else:
            raise PdbxError("Miscellaneous parsing error at line %d" % self.__curLineNumber)

    def __allSelected(self, container, catSelectD):
        """ Test the input container for completeness relative to the input category selection dictionary.
        """
        nl = -1
        if catSelectD:
            try:
                nl = container.getObjNameList()
                if len(nl) <= len(catSelectD):
                    ok = False
                else:
                    ok = True
                    logger.debug("nl %d length catSelectD %d returning %r" % (len(nl), len(catSelectD), ok))
            except Exception:
                ok = False
        else:
            ok = False
        return ok

    def __syntaxError(self, errText):
        msg = " [Line: %d] %s" % (self.__curLineNumber, errText)
        raise SyntaxError(msg)

    def __getContainerName(self, inWord):
        """ Returns the name of the data block or saveframe container
        """
        return str(inWord[5:]).strip()

    def __getState(self, inWord):
        """Identifies reserved syntax elements and assigns an associated state.

           on return: (reserved word, state)
           where -
              reserved word -  is one of CIF syntax elements:
                               data, loop, global, save, or stop
              state - the parser state required to process this next section.

        """
        i = inWord.find("_")
        if i == -1:
            return None, "ST_UNKNOWN"

        try:
            rWord = inWord[:i].lower()
            return rWord, self.__stateDict[rWord]
        except:
            return None, "ST_UNKNOWN"

    def __parser(self, tokenizer, containerList, categorySelectionD=None, excludeFlag=False):
        """ Parser for PDBx data files and dictionaries.

            Input - tokenizer() reentrant method recognizing data item names (_category.attribute)
                    quoted strings (single, double and multi-line semi-colon delimited), and unquoted
                    strings.

                    containerList -  list-type container for data and definition objects parsed from
                                     from the input file.

            On return:
                    The input containerList is appended with data and definition objects -
        """
        catSelectD = categorySelectionD if categorySelectionD is not None else {}
        logger.debug("Exclude Flag %r Category selection %r" % (excludeFlag, catSelectD))
        # Working container - data or definition
        curContainer = None
        # the last container of type data -
        previousDataContainer = None
        #
        # Working category container
        categoryIndex = {}
        curCategory = None
        #
        curRow = None
        state = None

        # Find the first reserved word and begin capturing data.
        #
        while True:
            curCatName, curAttName, curQuotedString, curWord = next(tokenizer)
            if curWord is None:
                continue
            reservedWord, state = self.__getState(curWord)
            if reservedWord is not None:
                break

        while True:
            #
            #  Set the current state  -
            #
            #  At this point in the processing cycle we are expecting a token containing
            #  either a '_category.attribute'  or a reserved word.
            #
            if curCatName is not None:
                state = "ST_KEY_VALUE_PAIR"
            elif curWord is not None:
                reservedWord, state = self.__getState(curWord)
            else:
                self.__syntaxError("Miscellaneous syntax error")
                return

            #
            # Process  _category.attribute  value assignments
            #
            if state == "ST_KEY_VALUE_PAIR":
                try:
                    curCategory = categoryIndex[curCatName]
                except KeyError:
                    # A new category is encountered - create a container and add a row
                    curCategory = categoryIndex[curCatName] = DataCategory(curCatName)
                    #
                    #  check if we have all of the selection
                    if not excludeFlag and self.__allSelected(curContainer, catSelectD):
                        return
                    try:
                        if catSelectD:
                            if not excludeFlag and curCatName in catSelectD:
                                curContainer.append(curCategory)
                            elif excludeFlag and curCatName not in catSelectD:
                                curContainer.append(curCategory)
                            else:
                                logger.debug("Skipped unselected/excluded category %s" % curCatName)
                        else:
                            curContainer.append(curCategory)
                    except AttributeError:
                        self.__syntaxError("Category cannot be added to  data_ block")
                        return

                    curRow = []
                    curCategory.append(curRow)
                else:
                    # Recover the existing row from the category
                    try:
                        # curRow = curCategory[0]
                        curRow = curCategory.getRow(0)
                    except IndexError:
                        self.__syntaxError("Internal index error accessing category data")
                        return

                # Check for duplicate attributes and add attribute to table.
                if curAttName in curCategory.getAttributeList():
                    self.__syntaxError("Duplicate attribute encountered in category")
                    return
                else:
                    curCategory.appendAttribute(curAttName)

                # Get the data for this attribute from the next token
                tCat, tAtt, curQuotedString, curWord = next(tokenizer)

                if tCat is not None or (curQuotedString is None and curWord is None):
                    self.__syntaxError("Missing data for item _%s.%s" % (curCatName, curAttName))

                if curWord is not None:
                    #
                    # Validation check token for misplaced reserved words  -
                    #
                    reservedWord, state = self.__getState(curWord)
                    if reservedWord is not None:
                        self.__syntaxError("Unexpected reserved word: %s" % (reservedWord))

                    curRow.append(curWord)

                elif curQuotedString is not None:
                    curRow.append(curQuotedString)

                else:
                    self.__syntaxError("Missing value in item-value pair")

                curCatName, curAttName, curQuotedString, curWord = next(tokenizer)
                continue

            #
            # Process a loop_ declaration and associated data -
            #
            elif state == "ST_TABLE":

                # The category name in the next curCatName,curAttName pair
                #    defines the name of the category container.
                curCatName, curAttName, curQuotedString, curWord = next(tokenizer)

                if curCatName is None or curAttName is None:
                    self.__syntaxError("Unexpected token in loop_ declaration")
                    return

                # Check for a previous category declaration.
                if curCatName in categoryIndex:
                    self.__syntaxError("Duplicate category declaration in loop_")
                    return

                curCategory = DataCategory(curCatName)

                #
                #  check if we have all of the selection
                if not excludeFlag and self.__allSelected(curContainer, catSelectD):
                    return
                try:
                    if catSelectD:
                        if not excludeFlag and curCatName in catSelectD:
                            curContainer.append(curCategory)
                        elif excludeFlag and curCatName not in catSelectD:
                            curContainer.append(curCategory)
                        else:
                            logger.debug("Skipped unselected/excluded category %s" % curCatName)
                    else:
                        curContainer.append(curCategory)
                except AttributeError:
                    self.__syntaxError("loop_ declaration outside of data_ block or save_ frame")
                    return

                curCategory.appendAttribute(curAttName)

                # Read the rest of the loop_ declaration
                while True:
                    curCatName, curAttName, curQuotedString, curWord = next(tokenizer)

                    if curCatName is None:
                        break

                    if curCatName != curCategory.getName():
                        self.__syntaxError("Changed category name in loop_ declaration")
                        return

                    curCategory.appendAttribute(curAttName)

                # If the next token is a 'word', check it for any reserved words -
                if curWord is not None:
                    reservedWord, state = self.__getState(curWord)
                    if reservedWord is not None:
                        if reservedWord == "stop":
                            return
                        else:
                            self.__syntaxError("Unexpected reserved word after loop declaration: %s" % (reservedWord))

                # Read the table of data for this loop_ -
                while True:
                    curRow = []
                    curCategory.append(curRow)

                    for tAtt in curCategory.getAttributeList():
                        if curWord is not None:
                            curRow.append(curWord)
                        elif curQuotedString is not None:
                            curRow.append(curQuotedString)

                        curCatName, curAttName, curQuotedString, curWord = next(tokenizer)

                    # loop_ data processing ends if -

                    # A new _category.attribute is encountered
                    if curCatName is not None:
                        break

                    # A reserved word is encountered
                    if curWord is not None:
                        reservedWord, state = self.__getState(curWord)
                        if reservedWord is not None:
                            break

                continue

            elif state == "ST_DEFINITION":
                # Ignore trailing unnamed saveframe delimiters e.g. 'save'
                sName = self.__getContainerName(curWord)
                if (len(sName) > 0):
                    curContainer = DefinitionContainer(sName)
                    containerList.append(curContainer)
                    categoryIndex = {}
                    curCategory = None
                else:
                    # reset current container to the last data contatiner
                    curContainer = previousDataContainer

                curCatName, curAttName, curQuotedString, curWord = next(tokenizer)

            elif state == "ST_DATA_CONTAINER":
                #
                dName = self.__getContainerName(curWord)
                if len(dName) == 0:
                    dName = "unidentified"
                curContainer = DataContainer(dName)
                containerList.append(curContainer)
                categoryIndex = {}
                curCategory = None
                previousDataContainer = curContainer
                curCatName, curAttName, curQuotedString, curWord = next(tokenizer)

            elif state == "ST_STOP":
                ###
                # curCatName, curAttName, curQuotedString, curWord = tokenizer.next()
                continue

            elif state == "ST_GLOBAL":
                curContainer = DataContainer("blank-global")
                curContainer.setGlobal()
                containerList.append(curContainer)
                categoryIndex = {}
                curCategory = None
                curCatName, curAttName, curQuotedString, curWord = next(tokenizer)

            elif state == "ST_UNKNOWN":
                self.__syntaxError("Unrecogized syntax element: " + str(curWord))
                return

    def __tokenizer(self, ifh):
        """ Tokenizer method for the mmCIF syntax file -

            Each return/yield from this method returns information about
            the next token in the form of a tuple with the following structure.

            (category name, attribute name, quoted strings, words w/o quotes or white space)


        """
        #
        # Regex definition for mmCIF syntax - semi-colon delimited strings are handled
        #                                     outside of this regex.
        #  Differentiated the reqular expression to the better handle embedded quotes.
        #
        mmcifRe = re.compile(
            r"(?:"

            "(?:_(.+?)[.](\S+))"               "|"  # _category.attribute

            "(?:['](.*?)(?:[']\s|[']$))"       "|"  # single quoted strings
            "(?:[\"](.*?)(?:[\"]\s|[\"]$))"    "|"  # double quoted strings

            "(?:\s*#.*$)"                      "|"  # comments (dumped)

            "(\S+)"                                 # unquoted words

            ")")

        fileIter = iter(ifh)

        # Tokenizer loop begins here ---
        while True:
            try:
                line = next(fileIter)
                self.__curLineNumber += 1

                # Dump comments
                if line.startswith("#"):
                    continue

                # Gobble up the entire semi-colon/multi-line delimited string and
                #    and stuff this into the string slot in the return tuple
                #
                if line.startswith(";"):
                    mlString = [line[1:]]
                    while True:
                        line = next(fileIter)
                        self.__curLineNumber += 1
                        if line.startswith(";"):
                            break
                        mlString.append(line)

                    # remove trailing new-line that is part of the \n; delimiter
                    mlString[-1] = mlString[-1].rstrip()
                    #
                    yield (None, None, "".join(mlString), None)
                    #
                    # Need to process the remainder of the current line -
                    line = line[1:]
                    # continue

                # Apply regex to the current line consolidate the single/double
                # quoted within the quoted string category
                for it in mmcifRe.finditer(line):
                    tgroups = it.groups()
                    #
                    if tgroups[4] is not None and tgroups[4].lower() == 'stop_':
                        continue
                    if tgroups != (None, None, None, None, None):
                        if tgroups[2] is not None:
                            qs = tgroups[2]
                        elif tgroups[3] is not None:
                            qs = tgroups[3]
                        else:
                            qs = None
                        groups = (tgroups[0], tgroups[1], qs, tgroups[4])
                        yield groups
            except StopIteration:
                return



##
# File:      MethodUtils.py
# Orignal:   Aug 12, 2013   Jdw
#
# Updates:
#   01-Aug-2017   jdw migrate portions to public repo
##
"""
Utility classes for applying dictionary methods on PDBx/mmCIF data files.
"""
from __future__ import absolute_import

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import sys

# from mmcif.api.PdbxContainers import *
from mmcif.api.DataCategory import DataCategory
from mmcif.api.DictionaryApi import DictionaryApi

import logging
logger = logging.getLogger(__name__)


class MethodUtils(object):

    def __init__(self, dictContainerList, verbose=False):
        #
        self.__verbose = verbose
        # list of dictionary data & definition containers
        self.__dictContainerList = dictContainerList
        self.__dApi = DictionaryApi(containerList=self.__dictContainerList, consolidate=True, verbose=self.__verbose)
        #
        # Target data container list
        self.__dataContainerList = []
        #

    def setDataContainerList(self, dataContainerList):
        self.__dataContainerList = dataContainerList

    def getDataContainerList(self):
        return self.__dataContainerList

    def getDictionary(self):
        return self.__dApi

    def getMethods(self):
        return self.__dApi.getMethodIndex()

    def getMethod(self, id):
        return self.__dApi.getMethod(id)

    def invokeMethods(self, fh=sys.stdout):
        mI = self.__dApi.getMethodIndex()
        lenD = len(mI)
        i = 0
        for k, mRefL in mI.items():
            for mRef in mRefL:
                i += 1
                id = mRef.getId()
                type = mRef.getType()
                categoryName = mRef.getCategoryName()
                attributeName = mRef.getAttributeName()
                #
                logger.debug("\n")
                logger.debug("++++++++++++++++++--------------------\n")
                logger.debug("Invoking dictionary method on file object: %s (%d/%d)\n" % (k, i, lenD))
                logger.debug(" + Method id: %s\n" % id)
                logger.debug(" + Type:      %s\n" % type)
                logger.debug(" + Category:  %s\n" % categoryName)
                logger.debug(" + Attribute: %s\n" % attributeName)
                #
                if type == "datablock":
                    logger.debug("Invoke datablock method %s\n" % id)
                    # self.invokeDataBlockMethod(type,self.__dApi.getMethod(id))
                    # continue
                #
                for db in self.__dataContainerList:
                    if type == "category":
                        if not db.exists(categoryName):
                            dc = DataCategory(categoryName)
                            db.append(dc)
                        dObj = db.getObj(categoryName)
                        dObj.invokeCategoryMethod(type, self.__dApi.getMethod(id), db)
                    elif type == "attribute":
                        if not db.exists(categoryName):
                            dc = DataCategory(categoryName)
                            db.append(dc)
                        dObj = db.getObj(categoryName)
                        # logger.debug("invoke -  %r %r %r %r" % (attributeName, type, self.__dApi.getMethod(id), db))
                        dObj.invokeAttributeMethod(attributeName, type, self.__dApi.getMethod(id), db)
                    elif type == "datablock":
                        logger.debug("Invoke datablock method %s\n" % id)
                        db.invokeDataBlockMethod(type, self.__dApi.getMethod(id), db)
                    else:
                        pass

    def dumpMethods(self, fh=sys.stdout):
        self.__dApi.dumpMethods(fh)

    def dumpDictionary(self, fh=sys.stdout):
        lenD = len(self.__dictContainerList)
        fh.write("\n--------------------------------------------\n")
        fh.write("\n-----------DUMP DICTIONARY------------------\n")
        fh.write("Dictionary object list length is: %d\n" % lenD)
        i = 1
        for dObj in self.__dictContainerList:
            if len(dObj.getName()) > 0:
                fh.write("\n")
                fh.write("++++++++++++++++++--------------------\n")
                fh.write("Dumping dictionary object named: %s (%d/%d)\n" %
                         (dObj.getName(), i, lenD))
                dObj.printIt(fh)
            i += 1
    #

    def dumpDataFile(self, fh=sys.stdout):
        lenD = len(self.__dataContainerList)
        fh.write("\n--------------------------------------------\n")
        fh.write("\n-----------DUMP DATAFILE--------------------\n")
        fh.write("Data object list length is: %d\n" % lenD)
        i = 1
        for dObj in self.__dataContainerList:
            fh.write("\n")
            fh.write("++++++++++++++++++--------------------\n")
            fh.write("Dumping data file object named: %s (%d/%d)\n" %
                     (dObj.getName(), i, lenD))
            dObj.printIt(fh)
            i += 1



##
#
# File:     DataCategory.py
# Original: 02-Feb-2009   jdw
#
# Update:
#   14-Nov-2012   jdw refactoring
#   17-Nov-2012   jdw self._rowList becomes data
#   21-Feb-2013   jdw add selectIndices(), selectValueWhere and selectValuesWhere.
#   10-Apr-2013   jdw add convenience method getValueOrDefault()
#   20-Jul-2015   jdw add selectIndicesFromList()
#   01-Aug-2017   jdw migrate portions to public repo
##
"""

A collection of container classes supporting the PDBx/mmCIF storage model.

"""
from __future__ import absolute_import
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"


from six.moves import range
from six.moves import zip


import sys
import logging
logger = logging.getLogger(__name__)

from mmcif.api.DataCategoryBase import DataCategoryBase


class DataCategory(DataCategoryBase):
    """  Methods for creating, accessing, and formatting PDBx/mmCif data categories.
    """

    def __init__(self, name, attributeNameList=None, rowList=None, raiseExceptions=True, copyInputData=True):
        """Summary

        Args:
            name (TYPE): Description
            attributeNameList (None, optional): Description
            rowList (None, optional): Description
            raiseExceptions (bool, optional): Description
        """
        super(DataCategory, self).__init__(name, attributeNameList, rowList, raiseExceptions=raiseExceptions, copyInputData=copyInputData)
        #
        self.__verbose = False
        self._currentRowIndex = 0
        self.__currentAttribute = None
        #

    def setVerboseMode(self, bool):
        self.__verbose = bool

    def getCurrentAttribute(self):
        return self.__currentAttribute

    def getRowIndex(self):
        return self._currentRowIndex

    def getFullRow(self, index):
        """ Return a full row based on the length of the the attribute list or a row initialized with missing values
        """
        try:
            if (len(self.data[index]) < self._numAttributes):
                for ii in range(self._numAttributes - len(self.data[index])):
                    self.data[index].append('?')
            return self.data[index]
        except Exception as e:
            return ['?' for ii in range(self._numAttributes)]

    def getAttributeListWithOrder(self):
        oL = []
        for ii, att in enumerate(self._attributeNameList):
            oL.append((att, ii))
        return oL

    def appendAttributeExtendRows(self, attributeName, defaultValue="?"):
        attributeNameLC = attributeName.lower()
        if attributeNameLC in self._catalog:
            i = self._attributeNameList.index(self._catalog[attributeNameLC])
            self._attributeNameList[i] = attributeName
            self._catalog[attributeNameLC] = attributeName
            logger.info("Appending existing attribute %s\n" % attributeName)
        else:
            self._attributeNameList.append(attributeName)
            self._catalog[attributeNameLC] = attributeName
            # add a placeholder to any existing rows for the new attribute.
            if (len(self.data) > 0):
                for row in self.data:
                    row.append(defaultValue)
            #
        self._numAttributes = len(self._attributeNameList)
        return self._numAttributes

    def getValue(self, attributeName=None, rowIndex=None):
        if attributeName is None:
            attribute = self.__currentAttribute
        else:
            attribute = attributeName
        if rowIndex is None:
            rowI = self._currentRowIndex
        else:
            rowI = rowIndex

        if isinstance(attribute, str) and isinstance(rowI, int):
            try:
                return self.data[rowI][self._attributeNameList.index(attribute)]
            except (IndexError):
                raise IndexError
        raise IndexError(attribute)

    def getValueOrDefault(self, attributeName=None, rowIndex=None, defaultValue=''):
        """  Within the current category return the value of input attribute in the input rowIndex [0-based].

             On error or if the value missing or null return the default value. Empty values returned as is.

             Exceptions on for unknown attributeNames and out-of-range indices.
        """
        if attributeName is None:
            attribute = self.__currentAttribute
        else:
            attribute = attributeName
        if rowIndex is None:
            rowI = self._currentRowIndex
        else:
            rowI = rowIndex

        if isinstance(attribute, self._string_types) and isinstance(rowI, int):
            try:
                tV = self.data[rowI][self._attributeNameList.index(attribute)]
                if ((tV is None) or (tV in ['.', '?'])):
                    return defaultValue
                else:
                    return tV
            except Exception as e:
                # logger.exception("Failing attributeName %s rowIndex %r defaultValue %r" % (attributeName, rowIndex, defaultValue))
                raise e
        else:
            raise ValueError
        #
        return defaultValue

    def getFirstValueOrDefault(self, attributeNameList, rowIndex=0, defaultValue=''):
        """ Return the value from the first non-null attribute found in the input attribute list
            from the row (rowIndex) in the current category object.
        """
        try:
            for at in attributeNameList:
                if self.hasAttribute(at):
                    tV = self.getValue(at, rowIndex)
                    if ((tV is None) or (tV in ['', '.', '?'])):
                        continue
                    else:
                        return tV
        except Exception as e:
            raise e

        return defaultValue

    def setValue(self, value, attributeName=None, rowIndex=None):
        """ Set the value of an existing attribute.  rowIndex values >=0, where
            the category will be extended in length as needed.
        """
        if attributeName is None:
            attribute = self.__currentAttribute
        else:
            attribute = attributeName

        if rowIndex is None:
            rowI = self._currentRowIndex
        else:
            rowI = rowIndex

        if isinstance(attribute, self._string_types) and isinstance(rowI, int) and (rowI >= 0):
            try:
                ind = -2
                # if row index is out of range - add the rows -
                for ii in range(rowI + 1 - len(self.data)):
                    self.data.append(self.__emptyRow())
                # self.data[rowI][attribute]=value
                ll = len(self.data[rowI])
                ind = self._attributeNameList.index(attribute)

                # extend the list if needed -
                if (ind >= ll):
                    self.data[rowI].extend([None for ii in range(ind - (ll - 1))])

                self.data[rowI][ind] = value
                return True
            except (IndexError):
                if self.__verbose:
                    logger.exception("DataCategory(setvalue) index error category %s attribute %s row index %d col %d rowlen %d value %r\n" %
                                     (self._name, attribute, rowI, ind, len(self.data[rowI]), value))
                    logger.debug("DataCategory(setvalue) attribute %r length attribute list %d \n" % (attribute, len(self._attributeNameList)))
                    for ii, a in enumerate(self._attributeNameList):
                        logger.debug("DataCategory(setvalue) %d attributeName %r\n" % (ii, a))

                raise IndexError
            except (ValueError):
                if self.__verbose:
                    logger.exception("DataCategory(setvalue) value error category %s attribute %s row index %d value %r\n" %
                                     (self._name, attribute, rowI, value))
                raise ValueError
        else:
            raise ValueError

    def __emptyRow(self):
        return [None for ii in range(len(self._attributeNameList))]

    def replaceValue(self, oldValue, newValue, attributeName):
        try:
            numReplace = 0
            if attributeName not in self._attributeNameList:
                return numReplace
            ind = self._attributeNameList.index(attributeName)
            for row in self.data:
                if row[ind] == oldValue:
                    row[ind] = newValue
                    numReplace += 1
            return numReplace
        except Exception as e:
            raise e

    def replaceSubstring(self, oldValue, newValue, attributeName):
        try:
            numReplace = 0
            if attributeName not in self._attributeNameList:
                return numReplace
            ind = self._attributeNameList.index(attributeName)
            for row in self.data:
                val = row[ind]
                row[ind] = val.replace(oldValue, newValue)
                if val != row[ind]:
                    numReplace += 1
            return numReplace
        except Exception as e:
            raise e

    def selectIndices(self, attributeValue, attributeName):
        try:
            rL = []
            if attributeName not in self._attributeNameList:
                return rL
            ind = self._attributeNameList.index(attributeName)
            for ii, row in enumerate(self.data):
                if attributeValue == row[ind]:
                    rL.append(ii)
            return rL
        except Exception as e:
            raise e

    def selectIndicesFromList(self, attributeValueList, attributeNameList):
        rL = []
        try:
            indList = []
            for at in attributeNameList:
                indList.append(self._attributeNameList.index(at))
            indValList = list(zip(indList, attributeValueList))
            #
            numList = len(indValList)
            for ii, row in enumerate(self.data):
                nMatch = 0
                for ind, tVal in indValList:
                    if tVal == row[ind]:
                        nMatch += 1
                if nMatch == numList:
                    rL.append(ii)
        except Exception as e:
            if self.__verbose:
                logger.exception("Selection/index failure for values %r" % attributeValueList)
            raise e

        return rL

    def selectValuesWhere(self, attributeName, attributeValueWhere, attributeNameWhere):
        rL = []
        try:
            ind = self._attributeNameList.index(attributeName)
            indWhere = self._attributeNameList.index(attributeNameWhere)
            for ii, row in enumerate(self.data):
                if attributeValueWhere == row[indWhere]:
                    rL.append(row[ind])
        except Exception as e:
            if self.__verbose:
                logger.exception("Selection failure")
            raise e
        return rL

    def selectValueListWhere(self, attributeNameList, attributeValueWhere, attributeNameWhere):
        """ Return a  list of lists containing the values of input attributeNameList
            satisfiying the attribute value where condition.
        """
        rL = []
        try:
            indList = []
            for at in attributeNameList:
                indList.append(self._attributeNameList.index(at))
            indWhere = self._attributeNameList.index(attributeNameWhere)
            for ii, row in enumerate(self.data):
                if attributeValueWhere == row[indWhere]:
                    rL.append([row[jj] for jj in indList])
        except Exception as e:
            if self.__verbose:
                logger.exception("Selection failure")
            raise e
        return rL

    def invokeAttributeMethod(self, attributeName, type, method, db):
        self._currentRowIndex = 0
        self.__currentAttribute = attributeName
        self.appendAttribute(attributeName)
        currentRowIndex = self._currentRowIndex
        #
        ind = self._attributeNameList.index(attributeName)
        if len(self.data) == 0:
            row = [None for ii in range(len(self._attributeNameList) * 2)]
            row[ind] = None
            self.data.append(row)

        for row in self.data:
            ll = len(row)
            if (ind >= ll):
                row.extend([None for ii in range(2 * ind - ll)])
                row[ind] = None
            exec(method.getInline(), globals(), locals())
            self._currentRowIndex += 1
            currentRowIndex = self._currentRowIndex

    def invokeCategoryMethod(self, type, method, db):
        self._currentRowIndex = 0
        exec(method.getInline(), globals(), locals())

    def printIt(self, fh=sys.stdout):
        fh.write("--------------------------------------------\n")
        fh.write("  Category: %s attribute list length: %d\n" %
                 (self._name, len(self._attributeNameList)))
        for at in self._attributeNameList:
            fh.write("  Category: %s attribute: %s\n" % (self._name, at))

        fh.write("  Row value list length: %d\n" % len(self.data))
        #
        for row in self.data[:2]:
            #
            if len(row) == len(self._attributeNameList):
                for ii, v in enumerate(row):
                    fh.write("        %30s: %s ...\n" % (self._attributeNameList[ii], str(v)[:30]))
            else:
                fh.write("+WARNING - %s data length %d attribute name length %s mismatched\n" %
                         (self._name, len(row), len(self._attributeNameList)))

    def dumpIt(self, fh=sys.stdout):
        fh.write("--------------------------------------------\n")
        fh.write("  Category: %s attribute list length: %d\n" %
                 (self._name, len(self._attributeNameList)))
        for at in self._attributeNameList:
            fh.write("  Category: %s attribute: %s\n" % (self._name, at))

        fh.write("  Value list length: %d\n" % len(self.data))
        for row in self.data:
            for ii, v in enumerate(row):
                fh.write("        %30s: %s\n" % (self._attributeNameList[ii], v))
##
##
